from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from ..client import Client, Command, CommandArg, CommandArgs
from ..models import Options
from ..responses import (
    BoundsNeSwResponses,
    CountResponse,
    HashesResponse,
    IdsResponse,
    ObjectsResponse,
    PointsResponse,
)
from .executable import Compiled, Executable
from .whereable import Whereable

Format = Literal["BOUNDS", "COUNT", "HASHES", "IDS", "OBJECTS", "POINTS"]
Output = Sequence[Format | int]


class Scan(Executable, Whereable):
    _key: str
    _command: Literal["SCAN"]
    _options: Options
    _output: Output | None = None
    _all: bool = False

    def __init__(self, client: Client, key: str) -> None:
        super().__init__(client)

        self.key(key)
        self._options = {}
        self._where = []
        self._wherein = []

    def key(self, key: str) -> Scan:
        self._key = key

        return self

    def cursor(self, value: int) -> Scan:
        self._options["cursor"] = value

        return self

    def limit(self, value: int) -> Scan:
        self._options["limit"] = value

        return self

    def nofields(self, flag: bool = True) -> Scan:
        self._options["nofields"] = flag

        return self

    def match(self, value: str) -> Scan:
        self._options["match"] = value

        return self

    def sparse(self, value: int) -> Scan:
        self._options["sparse"] = value

        return self

    def asc(self, flag: bool = True) -> Scan:
        self._options["asc"] = flag

        if flag:
            self._options["desc"] = False

        return self

    def desc(self, flag: bool = True) -> Scan:
        self._options["desc"] = flag

        if flag:
            self._options["asc"] = False

        return self

    def output(self, fmt: Format, precision: int | None = None) -> Scan:
        if fmt == "OBJECTS":
            self._output = None
        elif fmt == "HASHES" and precision:
            self._output = [fmt, precision]
        elif fmt == "BOUNDS" or fmt == "COUNT" or fmt == "IDS" or fmt == "POINTS":
            self._output = [fmt]

        return self

    async def asObjects(self) -> ObjectsResponse:
        self.output("OBJECTS")

        return ObjectsResponse(**(await self.exec()))

    async def asBounds(self) -> BoundsNeSwResponses:
        self.output("BOUNDS")

        return BoundsNeSwResponses(**(await self.exec()))

    async def asHashes(self, precision: int) -> HashesResponse:
        self.output("HASHES", precision)

        return HashesResponse(**(await self.exec()))

    async def asCount(self) -> CountResponse:
        self.output("COUNT")

        return CountResponse(**(await self.exec()))

    async def asIds(self) -> IdsResponse:
        self.output("IDS")

        return IdsResponse(**(await self.exec()))

    async def asPoints(self) -> PointsResponse:
        self.output("POINTS")

        return PointsResponse(**(await self.exec()))

    def __compile_options(self) -> CommandArgs:
        commands: list[CommandArg] = []

        # raises mypy: TypedDict key must be string literal
        # open PR: https://github.com/python/mypy/issues/7867
        for k in self._options:
            if isinstance(self._options[k], bool):  # type: ignore
                if self._options[k]:  # type: ignore
                    commands.append(k.upper())
            elif self._options[k] or self._options[k] == 0:  # type: ignore
                commands.extend([k.upper(), self._options[k]])  # type: ignore

        return commands

    def compile(self) -> Compiled:
        return [
            Command.SCAN.value,
            [
                self._key,
                *(self.__compile_options()),
                *(self.compile_where()),
                *(self.compile_wherein()),
                *(self._output if self._output else []),
            ],
        ]  # type: ignore



import random
import uuid
from typing import Any

Position = list[float]
Polygon = list[list[Position]]
BoundingBox = list[float]


def random_string() -> str:
    return str(uuid.uuid4())


def random_integer(minimum: int = 0, maximum: int = 1000) -> int:
    return random.randint(minimum, maximum)  # noqa: S311


def random_float(minimum: float = 0, maximum: float = 1000) -> float:
    return random.uniform(minimum, maximum)  # noqa: S311


def random_longitude(minimum: float = -179.0, maximum: float = 179.0) -> float:
    return random.uniform(minimum, maximum)  # noqa: S311


def random_latitude(minimum: float = -89.0, maximum: float = 89.0) -> float:
    return random.uniform(minimum, maximum)  # noqa: S311


def random_position() -> Position:
    return [random_longitude(), random_latitude()]


def random_point_geometry(
    position: Position | None = None,
) -> dict[str, str | list[float]]:
    return {"type": "Point", "coordinates": position if position else random_position()}


def random_polygon_geometry(
    bbox: BoundingBox = [13.1632, 52.4099, 13.6402, 52.6353],
) -> dict[str, str | Polygon]:
    [min_lng, min_lat, max_lng, max_lat] = bbox

    positions = [
        [random_longitude(min_lng, max_lng), random_latitude(min_lat, max_lat)]
        for _ in range(5)
    ]

    sw: list[float] = []
    ne: list[float] = []

    for [lng, lat] in positions:
        if len(sw) == 0:
            sw = [lng, lat]

        if len(ne) == 0:
            ne = [lng, lat]

        if sw[0] > lng:
            sw[0] = lng

        if sw[1] > lat:
            sw[1] = lat

        if ne[0] < lng:
            ne[0] = lng

        if ne[1] < lat:
            ne[1] = lat

    return {
        "type": "Polygon",
        "coordinates": [[sw, [ne[0], sw[1]], ne, [sw[0], ne[1]], sw]],
    }


def random_point_feature(
    geometry: dict[str, Any] = {}, properties: dict[str, Any] = {}
) -> dict:
    return {
        "type": "Feature",
        "geometry": geometry if geometry else random_point_geometry(),
        "properties": properties,
    }


def random_polygon_feature(
    geometry: dict[str, Any] = {}, properties: dict[str, Any] = {}
) -> dict:
    return {
        "type": "Feature",
        "geometry": geometry if geometry else random_polygon_geometry(),
        "properties": properties,
    }



from __future__ import annotations

import json
from typing import Any, Literal

from ..client import Client, Command
from ..responses import Fields, JSONResponse
from .executable import Compiled, Executable


class FSet(Executable):
    """Set the value for one or more fields of an id.

    Example::

        key = fleet
        id = truck1
        fields = {"driver":"John"}
        await tile38.set(key, id).point(52.25, 13.37).exec()
        await tile38.fset(key, id, fields)
    """

    _key: str
    _id: str
    _xx: Literal["XX"] | None = None
    _fields: Fields

    def __init__(self, client: Client, key: str, oid: str, fields: Fields) -> None:
        super().__init__(client)

        self._fields = {}
        self.key(key).id(oid).fields(fields)

    def key(self, key: str) -> FSet:
        """Set key of id to add fields to

        Args:
        key (str): key of a collection

        Returns:
            Self
        """
        self._key = key

        return self

    def id(self, oid: str) -> FSet:
        """Set id to add fields to

        Args:
        id (str): id of an object

        Returns:
            Self
        """
        self._id = oid

        return self

    def fields(self, fields: Fields) -> FSet:
        """Set fields to add to object in collection

        Args:
        fields (Fields): Fields to set

        Returns:
            Self
        """
        self._fields = fields

        return self

    def xx(self, flag: bool = True) -> FSet:
        """Alter behaviour if key/id does not exists

        If set, command will return 0 if key/id was not foundotherwise errors

        Args:
        flag (bool): Set True to return 0 if
        key/id not found else False to error on key/id not found.

        Returns:
            Self
        """
        self._xx = "XX" if flag else None

        return self

    @staticmethod
    def __unpack_fields(fields: Fields) -> list[Any]:
        command = []
        for k, v in fields.items():
            if isinstance(v, dict):
                command.extend([k, json.dumps(v)])
                continue

            command.extend([k, v])

        return command

    def compile(self) -> Compiled:
        return [
            Command.FSET.value,
            [
                self._key,
                self._id,
                *([self._xx] if self._xx else []),
                *(FSet.__unpack_fields(self._fields) if self._fields else []),
            ],
        ]  # type: ignore

    async def exec(self) -> JSONResponse:  # type: ignore[override]
        return JSONResponse(**(await self.client.command(*self.compile())))



from typing import Literal

from .client import Command, CommandArg
from .commands.fset import FSet
from .commands.set import Set
from .commands.setchan import SetChan
from .commands.sethook import SetHook
from .follower import Follower
from .responses import (
    Fields,
    InfoLeaderResponse,
    JSONResponse,
    ServerStatsResponseLeader,
    TTLResponse,
)


class Leader(Follower):
    """
    A class to interact with a Tile38 leader instance, extending the Follower class
    with additional write and administrative commands.

    Attributes:
        client (Client): An instance of the Client class for connecting to Tile38.
    """

    async def delete(self, key: str, oid: str) -> JSONResponse:
        """Delete an object from a collection.

        Tile38 Command:
            DEL key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            JSONResponse: Confirmation of deletion.
        """
        return JSONResponse(**(await self.client.command(Command.DEL, [key, oid])))

    async def delchan(self, name: str) -> JSONResponse:
        """Delete a channel.

        Tile38 Command:
            DELCHAN name

        Args:
            name (str): The channel name.

        Returns:
            JSONResponse: Confirmation of channel deletion.
        """
        return JSONResponse(**(await self.client.command(Command.DELCHAN, [name])))

    async def delhook(self, name: str) -> JSONResponse:
        """Delete a geofence hook.

        Tile38 Command:
            DELHOOK name

        Args:
            name (str): The hook name.

        Returns:
            JSONResponse: Confirmation of hook deletion.
        """
        return JSONResponse(**(await self.client.command(Command.DELHOOK, [name])))

    async def drop(self, key: str) -> JSONResponse:
        """Drop an entire collection.

        Tile38 Command:
            DROP key

        Args:
            key (str): The collection key.

        Returns:
            JSONResponse: Confirmation of collection drop.
        """
        return JSONResponse(**(await self.client.command(Command.DROP, [key])))

    async def expire(self, key: str, oid: str, seconds: int) -> JSONResponse:
        """Set an expiration time for an object.

        Tile38 Command:
            EXPIRE key id seconds

        Args:
            key (str): The collection key.
            id (str): The object ID.
            seconds (int): The expiration time in seconds.

        Returns:
            JSONResponse: Confirmation of expiration set.
        """
        # TODO: fix mypy
        # for reasons unknown [key, id, seconds] has type List[object]
        # and fails mypy validation
        p: list[CommandArg] = [key, oid, seconds]
        response = await self.client.command(Command.EXPIRE, p)

        return JSONResponse(**response)

    async def flushdb(self) -> JSONResponse:
        """Delete all data from the Tile38 database.

        Tile38 Command:
            FLUSHDB

        Returns:
            JSONResponse: Confirmation of database flush.
        """
        return JSONResponse(**(await self.client.command(Command.FLUSHDB)))

    def fset(self, key: str, oid: str, fields: Fields) -> FSet:
        """Set fields on an existing object.

        Tile38 Command:
            FSET key id field value...

        Args:
            key (str): The collection key.
            id (str): The object ID.
            fields (Fields): The fields to set.

        Returns:
            Fset: An object to perform field set operations.
        """
        return FSet(self.client, key, oid, fields)

    # TODO: how can I override supertype Follower here correctly?
    async def info(self) -> InfoLeaderResponse:  # type: ignore[override]
        """Get detailed server information specific to the leader.

        Tile38 Command:
            INFO

        Returns:
            InfoLeaderResponse: Server information and statistics.
        """
        return InfoLeaderResponse(**(await self.client.command(Command.INFO)))

    async def jset(
        self,
        key: str,
        oid: str | int,
        path: str,
        value: str,
        mode: Literal["RAW", "STR"] | None = None,
    ) -> JSONResponse:
        """Set a JSON value in an object.

        Tile38 Command:
            JSET key id path value [RAW|STR]

        Args:
            key (str): The collection key.
            id (Union[str, int]): The object ID.
            path (str): The JSON path.
            value (str): The value to set.
            mode (Optional[Literal["RAW", "STR"]]): Mode for setting value.

        Returns:
            JSONResponse: Confirmation of JSON value set.
        """
        return JSONResponse(
            **(
                await self.client.command(
                    Command.JSET, [key, oid, path, value, *([mode] if mode else [])]
                )
            )
        )

    async def jdel(self, key: str, oid: str, path: str) -> JSONResponse:
        """Delete a JSON value from an object.

        Tile38 Command:
            JDEL key id path

        Args:
            key (str): The collection key.
            id (str): The object ID.
            path (str): The JSON path.

        Returns:
            JSONResponse: Confirmation of JSON value deletion.
        """
        return JSONResponse(
            **(await self.client.command(Command.JDEL, [key, oid, path]))
        )

    async def pdel(self, key: str, pattern: str) -> JSONResponse:
        """Delete objects matching a pattern.

        Tile38 Command:
            PDEL key pattern

        Args:
            key (str): The collection key.
            pattern (str): The pattern to match IDs.

        Returns:
            JSONResponse: Confirmation of pattern-based deletion.
        """
        return JSONResponse(**(await self.client.command(Command.PDEL, [key, pattern])))

    async def pdelchan(self, pattern: str) -> JSONResponse:
        """Delete channels matching a pattern.

        Tile38 Command:
            PDELCHAN pattern

        Args:
            pattern (str): The pattern to match channel names.

        Returns:
            JSONResponse: Confirmation of channel deletion.
        """
        return JSONResponse(**(await self.client.command(Command.PDELCHAN, [pattern])))

    async def pdelhook(self, pattern: str) -> JSONResponse:
        """Delete hook matching a pattern.

        Tile38 Command:
            PDELHOOK pattern

        Args:
            pattern (str): The pattern to match hook names.

        Returns:
            JSONResponse: Confirmation of hook deletion.
        """
        return JSONResponse(**(await self.client.command(Command.PDELHOOK, [pattern])))

    async def persist(self, key: str, oid: str) -> JSONResponse:
        """Remove expiration from an object.

        Tile38 Command:
            PERSIST key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            JSONResponse: Confirmation of persistence.
        """
        return JSONResponse(**(await self.client.command(Command.PERSIST, [key, oid])))

    async def readonly(self, value: bool = True) -> JSONResponse:
        """Set the server to readonly mode.

        Tile38 Command:
            READONLY yes|no

        Args:
            value (bool): True to enable readonly mode, False to disable.

        Returns:
            JSONResponse: Confirmation of readonly status.
        """
        return JSONResponse(
            **(await self.client.command(Command.READONLY, ["yes" if value else "no"]))
        )

    async def rename(self, key: str, newkey: str, nx: bool = False) -> JSONResponse:
        """Rename a collection key.

        Tile38 Command:
            RENAME key newkey
            RENAMENX key newkey

        Args:
            key (str): The current collection key.
            newkey (str): The new collection key.
            nx (bool): If True, rename only if the new key does not exist.

        Returns:
            JSONResponse: Confirmation of key renaming.
        """
        command = Command.RENAMENX if nx else Command.RENAME
        return JSONResponse(**(await self.client.command(command, [key, newkey])))

    def set(self, key: str, oid: str) -> Set:
        """Set an object in a collection.

        Tile38 Command:
            SET key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            Set: An object to perform set operations.
        """
        return Set(self.client, key, oid)

    # TODO: how can I override supertype Follower here correctly?
    async def server(self) -> ServerStatsResponseLeader:  # type: ignore[override]
        """Get server stats specific to the leader.

        Tile38 Command:
            SERVER

        Returns:
            ServerStatsResponseLeader: Server statistics.
        """
        return ServerStatsResponseLeader(**(await self.client.command(Command.SERVER)))

    def sethook(self, name: str, endpoint: str) -> SetHook:
        """Set a geofence hook.

        Tile38 Command:
            SETHOOK name endpoint

        Args:
            name (str): The hook name.
            endpoint (str): The endpoint URL.

        Returns:
            SetHook: An object to configure the hook.
        """
        return SetHook(self.client, name, endpoint)

    def setchan(self, name: str) -> SetChan:
        """Create a pub/sub channel.

        Tile38 Command:
            SETCHAN name

        Args:
            name (str): The channel name.

        Returns:
            SetChan: An object to configure the channel.
        """
        return SetChan(self.client, name)

    async def ttl(self, key: str, oid: str) -> TTLResponse:
        """Get the time-to-live (TTL) of an object.

        Tile38 Command:
            TTL key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            TTLResponse: The TTL in seconds.
        """
        return TTLResponse(**(await self.client.command(Command.TTL, [key, oid])))



import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_config(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    response = await tile38.config_set("keepalive", 500)
    assert response.ok

    response = await tile38.follower().config_set("keepalive", 500)
    assert response.ok

    response = await tile38.config_rewrite()
    assert response.ok
    response = await tile38.follower().config_rewrite()
    assert response.ok

    #  NOTE: tile38 does only return string types not sure if i should parse
    config_keepalive = await tile38.config_get("keepalive")
    assert config_keepalive.properties["keepalive"] == "500"
    config_keepalive = await tile38.follower().config_get("keepalive")
    assert config_keepalive.properties["keepalive"] == "500"



from __future__ import annotations

from typing import Any, Literal

from ..client import Client, Command, SubCommand
from ..commands.intersects import Intersects
from ..commands.nearby import Nearby
from ..commands.within import Within
from ..responses import Meta
from .executable import Compiled, Executable


class SetChan(Executable):
    _command: Literal["SETCHAN"] = "SETCHAN"
    _meta: Meta | None
    _ex: int | None

    def __init__(self, client: Client, name: str) -> None:
        super().__init__(client)

        self.name(name)
        self._meta = {}
        self._ex = None

    def name(self, name: str) -> SetChan:
        self._name = name

        return self

    def meta(self, meta: Meta) -> SetChan:
        self._meta = meta

        return self

    def ex(self, ex: int) -> SetChan:
        self._ex = ex

        return self

    def nearby(self, key: str) -> Nearby:
        return Nearby(self.client, key, self).fence()

    def within(self, key: str) -> Within:
        return Within(self.client, key, self).fence()

    def intersects(self, key: str) -> Intersects:
        return Intersects(self.client, key, self).fence()

    @staticmethod
    def __unpack_meta(meta: Meta) -> list[Any]:
        command = []
        for k, v in meta.items():
            command.extend([SubCommand.META.value, k, v])

        return command

    def compile(self) -> Compiled:
        return [
            Command.SETCHAN.value,
            [
                self._name,
                *(SetChan.__unpack_meta(self._meta) if self._meta else []),
                *([SubCommand.EX.value, self._ex] if self._ex else []),
            ],
        ]  # type: ignore



import pytest

from pyle38.errors import Tile38Error
from pyle38.parse_response import parse_response


def test_parse_response() -> None:
    with pytest.raises(Tile38Error):
        response = parse_response(1)
        assert response

    with pytest.raises(Tile38Error):
        response = parse_response("{},")
        assert response



from collections.abc import Generator
from typing import Any


def flatten(arr: list[list[Any]] | list[Any]) -> Generator:
    for x in arr:
        if hasattr(x, "__iter__") and not isinstance(x, str):
            yield from flatten(x)
        else:
            yield x



import os
from collections.abc import AsyncGenerator, Awaitable, Callable
from typing import Any

import pytest

from pyle38 import Tile38

default_tile38_leader_url = os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851"
default_tile38_follower_url = (
    os.getenv("TILE38_FOLLOWER_URI") or "redis://localhost:9852"
)


@pytest.fixture()
async def create_tile38():  # type: ignore[no-untyped-def]
    teardown_clients = []

    async def client_factory(url: str = default_tile38_leader_url) -> Tile38:
        tile38 = Tile38(url)
        # make sure to reset readonly
        await tile38.readonly(False)

        async def teardown() -> None:
            try:
                await tile38.flushdb()
                await tile38.readonly(False)
            # TODO: find explicit exception
            except Exception as e:
                print(e)
                await tile38.flushdb()

            await tile38.quit()

        teardown_clients.append(teardown)
        return tile38

    yield client_factory

    for teardown in teardown_clients:
        await teardown()


@pytest.fixture()
async def create_tile38_with_follower():  # type: ignore[no-untyped-def]
    teardown_clients = []

    async def client_factory(
        url: str = default_tile38_leader_url,
        follower_url: str = default_tile38_follower_url,
    ) -> Tile38:
        tile38 = Tile38(url, follower_url)
        # make sure to reset readonly
        await tile38.readonly(False)

        async def teardown() -> None:
            try:
                await tile38.flushdb()
            # TODO: find explicit exception
            except Exception as e:
                print(e)
                await tile38.flushdb()

            await tile38.quit()

        teardown_clients.append(teardown)
        return tile38

    yield client_factory

    for teardown in teardown_clients:
        await teardown()


@pytest.fixture()
async def tile38(
    create_tile38: Callable[..., Awaitable[Tile38]],
) -> AsyncGenerator[Any, Any]:
    yield await create_tile38()


@pytest.fixture()
async def tile38_with_follower(
    create_tile38_with_follower: Callable[..., Awaitable[Tile38]],
) -> AsyncGenerator[Any, Any]:
    yield await create_tile38_with_follower()



import time

import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_server(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    leaderResponse = await tile38.server()
    assert leaderResponse.ok
    assert leaderResponse.stats.num_points == 0

    with pytest.raises(AttributeError):
        # NOTE: we know that type is wrong here, we just wanna test it not being here
        assert leaderResponse.stats.caught_up  # type: ignore[attr-defined,unused-ignore]

    time.sleep(0.2)

    followerResponse = await tile38.follower().server()
    assert followerResponse.ok
    assert followerResponse.stats.num_points == 0
    assert followerResponse.stats.caught_up is True
    assert followerResponse.stats.caught_up_once is True



import pytest

from pyle38 import Tile38

from .helper.random_data import random_integer, random_polygon_geometry, random_string


@pytest.mark.asyncio
async def test_command_expire_persist(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    obj = random_polygon_geometry()
    expire = random_integer(10, 100)

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.ttl(key, oid)
    assert response.ok
    assert response.ttl == -1

    response = await tile38.expire(key, oid, expire)
    assert response.ok

    response = await tile38.ttl(key, oid)
    assert response.ok
    assert response.ttl < expire

    response = await tile38.persist(key, oid)
    assert response.ok

    response = await tile38.ttl(key, oid)
    assert response.ok
    assert response.ttl == -1



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from ..client import Client, Command, CommandArg, CommandArgs, SubCommand
from ..errors import Pyle38NoHookToActivateError
from ..models import Options, PointQuery
from ..responses import (
    BoundsNeSwResponses,
    CountResponse,
    FenceCommand,
    FenceDetect,
    HashesResponse,
    IdsResponse,
    JSONResponse,
    ObjectsResponse,
    PointsResponse,
)
from ..utils import flatten
from .executable import Compiled, Executable
from .whereable import Whereable

Format = Literal["BOUNDS", "COUNT", "HASHES", "IDS", "OBJECTS", "POINTS"]
Output = Sequence[Format | int]


class Nearby(Executable, Whereable):
    """Nearby searches a key for objects that are close by."""

    _key: str
    _command: Literal["NEARBY"]
    _hook = None
    _options: Options
    _query: PointQuery
    _output: Output | None = None
    _all: bool = False
    _fence: bool = False
    _detect: list[FenceDetect] | None
    _commands: list[FenceCommand] | None

    def __init__(
        self, client: Client, key: str, hook: Executable | None = None
    ) -> None:
        """__init__.

        Args:
            client (Client): client
            key (str): key to search in
            hook: Endpoint to send events to
                Options: AMQP, KAFKA, HTTPS/HTTP, SQS, MQTT, NATS, gRPC, Redis

        Returns:
            None
        """
        super().__init__(client)

        self.key(key)
        self._options = {}
        self._hook = hook
        self._detect = []
        self._where = []
        self._wherein = []
        self._commands = []

    def key(self, key: str) -> Nearby:
        """Set key to search in

        Args:
            key (str): key of a collection to search in

        Returns:
            Nearby
        """
        self._key = key

        return self

    def cursor(self, value: int) -> Nearby:
        """Cursor used to iterate though objects of the search results.

        Used for pagination through search results.

        Args:
            value (int): cursor value to start at, default: 0 if not set

        Returns:
            Nearby
        """
        self._options["cursor"] = value

        return self

    def fence(self, flag: bool = True) -> Nearby:
        """Flag to indicate that the Nearby Query is used for a geo fence.

        Args:
            flag (bool): defaults to True if not set

        Returns:
            Nearby
        """
        self._fence = flag

        return self

    def detect(self, what: list[FenceDetect]) -> Nearby:
        """Option to filter the type of events in a geo fence.

        Args:
            what (List[FenceDetect]): what to detect in geo fence events.
                defaults to 'enter,exit,outside,inside,crosses' if not set

        Returns:
            Nearby
        """
        self._detect = what if len(what) > 0 else []

        return self

    def commands(self, which: list[FenceCommand] | None = []) -> Nearby:
        """Option to filter what commands should be triggering a geo fence event.

        Args:
            which (Optional[List[FenceCommand]]): which commands trigger a geo fence
            event defaults to 'SET,DEL,JSET,JDEL' if not set

        Returns:
            Nearby
        """
        if which and len(which) > 0:
            self._commands = which

        return self

    def limit(self, value: int) -> Nearby:
        """Limit the number of returned objects in a search.

        Also used for pagination through search results.

        Args:
            value (int): limit value, defaults to 100 if not set

        Returns:
            Nearby
        """
        self._options["limit"] = value

        return self

    def nofields(self, flag: bool = True) -> Nearby:
        """Option to explicitly not return fields in search results.

        Args:
            flag (bool): flag

        Returns:
            Nearby
        """
        self._options["nofields"] = flag

        return self

    def match(self, value: str) -> Nearby:
        """Match can be used to filtered objects considered in the search with a
        glob pattern.

        Args:
            value (str): value

        Returns:
            Nearby
        """
        self._options["match"] = value

        return self

    def sparse(self, value: int) -> Nearby:
        """Instead of returning all results of a search. Return a sparse result evenly
        distributed in the given search area. EXPERIMENTAL

        Args:
            value (int): values between 1 and 8

        Returns:
            Nearby
        """
        self._options["sparse"] = value

        return self

    def distance(self, flag: bool = True) -> Nearby:
        """Return haversine distance to found objects

        Args:
            flag (bool): flag

        Returns:
            Nearby
        """
        self._options["distance"] = flag

        return self

    def point(self, lat: float, lon: float, radius: float | None = None) -> Nearby:
        """Define a point as input bounding area for the nearby search.

        Args:
            lat (float): center latitude of point
            lon (float): center longitude of point
            radius (float): radius of point

        Returns:
            Nearby
        """
        self._query = PointQuery(lat=lat, lon=lon, radius=radius)

        return self

    def output(self, fmt: Format, precision: int | None = None) -> Nearby:
        """Define an output format for query results.

        Args:
            format (Format): format,
                eg. 'OBJECTS', 'HASHES', 'BOUNDS', 'COUNT', 'IDS', 'POINTS'
            precision (Optional[int]): precision

        Returns:
            Nearby
        """
        if fmt == "OBJECTS":
            self._output = None
        elif fmt == "HASHES" and precision:
            self._output = [fmt, precision]
        elif fmt == "BOUNDS" or fmt == "COUNT" or fmt == "IDS" or fmt == "POINTS":
            self._output = [fmt]

        return self

    async def asObjects(self) -> ObjectsResponse:
        """Return query results as objects.

        Args:

        Returns:
            ObjectsResponse
        """
        self.output("OBJECTS")

        return ObjectsResponse(**(await self.exec()))

    async def asBounds(self) -> BoundsNeSwResponses:
        """Return query results as bounds.

        Args:

        Returns:
            BoundsNeSwResponses
        """
        self.output("BOUNDS")

        return BoundsNeSwResponses(**(await self.exec()))

    async def asHashes(self, precision: int) -> HashesResponse:
        """Return query results as geohashes.

        Args:
            precision (int): precision of the returned geohash

        Returns:
            HashesResponse
        """
        self.output("HASHES", precision)

        return HashesResponse(**(await self.exec()))

    async def asCount(self) -> CountResponse:
        """Return query results as count only.

        Args:

        Returns:
            CountResponse
        """
        self.output("COUNT")

        return CountResponse(**(await self.exec()))

    async def asIds(self) -> IdsResponse:
        """Return query results as object ids only.

        Args:

        Returns:
            IdsResponse
        """
        self.output("IDS")

        return IdsResponse(**(await self.exec()))

    async def asPoints(self) -> PointsResponse:
        """Return query results as points.

        Args:

        Returns:
            PointsResponse
        """
        self.output("POINTS")

        return PointsResponse(**(await self.exec()))

    def __compile_options(self) -> CommandArgs:
        """__compile_options.

        Args:

        Returns:
            CommandArgs
        """
        commands: list[CommandArg] = []

        # raises mypy: TypedDict key must be string literal
        # open PR: https://github.com/python/mypy/issues/7867
        for k in self._options:
            if isinstance(self._options[k], bool):  # type: ignore
                commands.append(k.upper())
            elif self._options[k] or self._options[k] == 0:  # type: ignore
                commands.extend([k.upper(), self._options[k]])  # type: ignore

        return commands

    def __compile_fence(self) -> CommandArgs:
        """__compile_fence

        Args:

        Returns:
            CommandArgs
        """
        return (
            [
                SubCommand.FENCE.value,
                *(
                    [SubCommand.DETECT.value, ",".join(self._detect)]
                    if self._detect
                    else []
                ),
                *(
                    [SubCommand.COMMANDS.value, ",".join(self._commands)]
                    if self._commands
                    else []
                ),
            ]
            if self._fence
            else []
        )

    def compile(self) -> Compiled:
        """compile

        Args:

        Returns:
            Compiled
        """
        compiled = [
            Command.NEARBY.value,
            [
                self._key,
                *(self.__compile_options()),
                *(self.compile_where()),
                *(self.compile_wherein()),
                *(self.__compile_fence()),
                *(self._output if self._output else []),
                *(self._query.get()),
            ],
        ]

        if self._hook:
            command, args = self._hook.compile()
            return [command, [*(args), *(flatten(compiled))]]  # type: ignore

        return compiled  # type: ignore

    async def activate(self) -> JSONResponse:
        """Activate is used in SetHook to activate a geo-fenced search.

        Args:

        Returns:
            JSONResponse
        """
        if self._hook:
            return JSONResponse(**(await self.client.command(*self.compile())))
        else:
            raise Pyle38NoHookToActivateError



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from ..client import Client, Command, CommandArg, CommandArgs, SubCommand
from ..errors import Pyle38BadObjectInputError, Pyle38NoHookToActivateError
from ..models import (
    BoundsQuery,
    CircleQuery,
    Feature,
    GetQuery,
    HashQuery,
    ObjectQuery,
    Options,
    Polygon,
    QuadkeyQuery,
    SectorQuery,
    TileQuery,
)
from ..responses import (
    BoundsNeSwResponses,
    CountResponse,
    FenceCommand,
    FenceDetect,
    HashesResponse,
    IdsResponse,
    JSONResponse,
    ObjectsResponse,
    PointsResponse,
)
from ..utils import flatten
from .executable import Compiled, Executable
from .whereable import Whereable

Format = Literal["BOUNDS", "COUNT", "HASHES", "IDS", "OBJECTS", "POINTS"]
Output = Sequence[Format | int]


class Within(Executable, Whereable):
    """WITHIN searches a key for objects that are fully contained
    within a given bounding area."""

    _key: str
    _command: Literal["WITHIN"]
    _hook = None
    _options: Options
    _query: (
        CircleQuery
        | BoundsQuery
        | HashQuery
        | QuadkeyQuery
        | TileQuery
        | ObjectQuery
        | GetQuery
        | SectorQuery
    )
    _output: Output | None = None
    _all: bool = False
    _fence: bool = False
    _detect: list[FenceDetect] | None
    _commands: list[FenceCommand] | None

    def __init__(
        self, client: Client, key: str, hook: Executable | None = None
    ) -> None:
        """__init__.

        Args:
            client (Client): client
            key (str): key to search in
            hook: Endpoint to send events to
                Options: AMQP, KAFKA, HTTPS/HTTP, SQS, MQTT, NATS, gRPC, Redis

        Returns:
            None
        """
        super().__init__(client)

        self.key(key)
        self._options = {}
        self._hook = hook
        self._where = []
        self._wherein = []
        self._detect = []
        self._commands = []

    def key(self, key: str) -> Within:
        """Set key to search in

        Args:
            key (str): key of a collection to search in

        Returns:
            Within
        """
        self._key = key

        return self

    def circle(self, lat: float, lon: float, radius: float) -> Within:
        """Define a circle as input bounding area for the within search.

        Args:
            lat (float): center latitude of circle
            lon (float): center longitude of circle
            radius (float): radius of circle

        Returns:
            Within
        """
        self._query = CircleQuery(lat=lat, lon=lon, radius=radius)

        return self

    def cursor(self, value: int) -> Within:
        """Cursor used to iterate though objects of the search results.

        Used for pagination through search results.

        Args:
            value (int): cursor value to start at, default: 0 if not set

        Returns:
            Within
        """
        self._options["cursor"] = value

        return self

    def buffer(self, value: int) -> Within:
        """Apply a buffer around area formats to increase the search area by x meters.

        Args:
            value (int): buffer size in meters

        Returns:
            Within
        """
        self._options["buffer"] = value

        return self

    def limit(self, value: int) -> Within:
        """Limit the number of returned objects in a search.

        Also used for pagination through search results.

        Args:
            value (int): limit value, defaults to 100 if not set

        Returns:
            Within
        """
        self._options["limit"] = value

        return self

    def fence(self, flag: bool = True) -> Within:
        """Flag to indicate that the Within Query is used for a geo fence.

        Args:
            flag (bool): defaults to True if not set

        Returns:
            Within
        """
        self._fence = flag

        return self

    def detect(self, what: list[FenceDetect]) -> Within:
        """Option to filter the type of events in a geo fence.

        Args:
            what (List[FenceDetect]): what to detect in geo fence events.
                defaults to 'enter,exit,outside,inside,crosses' if not set

        Returns:
            Within
        """
        self._detect = what if len(what) > 0 else []

        return self

    def commands(self, which: list[FenceCommand] | None = []) -> Within:
        """Option to filter what commands should be triggering a geo fence event.

        Args:
            which (Optional[List[FenceCommand]]): which commands trigger a geo fence
            event defaults to 'SET,DEL,JSET,JDEL' if not set

        Returns:
            Within
        """

        if which and len(which) > 0:
            self._commands = which

        return self

    def nofields(self, flag: bool = True) -> Within:
        """Option to explicitly not return fields in search results.

        Args:
            flag (bool): flag

        Returns:
            Within
        """
        self._options["nofields"] = flag

        return self

    def match(self, value: str) -> Within:
        """Match can be used to filtered objects considered in the search
        with a glob pattern.

        Args:
            value (str): value

        Returns:
            Within
        """
        self._options["match"] = value

        return self

    def sparse(self, value: int) -> Within:
        """Instead of returning all results of a search.
        Return a sparse result evenly distributed
        in the given search area. EXPERIMENTAL

        Args:
            value (int): values between 1 and 8

        Returns:
            Within
        """
        self._options["sparse"] = value

        return self

    def bounds(
        self, minlat: float, minlon: float, maxlat: float, maxlon: float
    ) -> Within:
        """Define a bounding as input bounding area for the within search.

        A bounding box is build from its most south-western and its most
        north-eastern point.

        Args:
            minlat (float): minimum latitude / south-west latitude
            minlon (float): minimum longitude / south-west longitude
            maxlat (float): maximum latitude / north-east latitude
            maxlon (float): maximum longitude / north-east longitude

        Returns:
            Within
        """
        self._query = BoundsQuery(
            minlat=minlat, minlon=minlon, maxlat=maxlat, maxlon=maxlon
        )

        return self

    def hash(self, geohash: str) -> Within:
        """Define a geohash as input bounding area for the within search.

        Args:
            geohash (str): geohash, eg. 'gcpvp'

        Returns:
            Within
        """
        self._query = HashQuery(geohash=geohash)

        return self

    def quadkey(self, quadkey: str) -> Within:
        """Define a quadkey as input bounding area for the within search.

        Args:
            quadkey (str): quadkey, eg. '120'

        Returns:
            Within
        """
        self._query = QuadkeyQuery(quadkey=quadkey)

        return self

    def sector(
        self, lat: float, lon: float, radius: float, bearing1: float, bearing2: float
    ) -> Within:
        """Define a sector as inbound area type for the within search

        Args:
            lat (float): latitude of sector center
            lon (float): longitude of sector center
            radius (float): radius of sector
            bearing1 (float): start bearing
            bearing2 (float): end bearing

        Returns:
            Within
        """
        self._query = SectorQuery(
            lat=lat, lon=lon, radius=radius, bearing1=bearing1, bearing2=bearing2
        )

        return self

    def tile(self, x: int, y: int, z: int) -> Within:
        """Define a tile as input bounding area for the within search.

        Args:
            x (int): x coordinate of the tile
            y (int): y coordinate of the tile
            z (int): z zoom level for the tile

        Returns:
            Within
        """
        self._query = TileQuery(x=x, y=y, z=z)

        return self

    def object(self, obj: Polygon | Feature | dict) -> Within:
        """Define an object as input bounding area for the intersects search.

        Args:
            object (Union[Polygon, Feature, Dict]): GeoJSON Feature or Polygon Geometry

        Returns:
            Intersects
        """
        if isinstance(obj, dict) and obj.get("type") == "Polygon":
            obj = Polygon(**obj)
            self._query = ObjectQuery(object=obj)
            return self

        if isinstance(obj, dict) and obj.get("type") == "Feature":
            obj = Feature(**obj)
            self._query = ObjectQuery(object=obj)
            return self

        if isinstance(obj, Polygon | Feature):
            self._query = ObjectQuery(object=obj)
            return self

        raise Pyle38BadObjectInputError()

    def get(self, key: str, oid: str) -> Within:
        """Define an object in a collection as bounding area for the within search.

        Args:
            key (str): key of the collection
            id (str): id of the object in the collection

        Returns:
            Within
        """
        self._query = GetQuery(key=key, id=oid)

        return self

    def output(self, fmt: Format, precision: int | None = None) -> Within:
        """Define an output format for query results.

        Args:
            format (Format): format,
                eg. 'OBJECTS', 'HASHES', 'BOUNDS', 'COUNT', 'IDS', 'POINTS'
            precision (Optional[int]): precision

        Returns:
            Within
        """
        if fmt == "OBJECTS":
            self._output = None
        elif fmt == "HASHES" and precision:
            self._output = [fmt, precision]
        elif fmt == "BOUNDS" or fmt == "COUNT" or fmt == "IDS" or fmt == "POINTS":
            self._output = [fmt]

        return self

    async def asObjects(self) -> ObjectsResponse:
        """Return query results as objects.

        Args:

        Returns:
            ObjectsResponse
        """
        self.output("OBJECTS")

        return ObjectsResponse(**(await self.exec()))

    async def asBounds(self) -> BoundsNeSwResponses:
        """Return query results as bounds.

        Args:

        Returns:
            BoundsNeSwResponses
        """
        self.output("BOUNDS")

        return BoundsNeSwResponses(**(await self.exec()))

    async def asHashes(self, precision: int) -> HashesResponse:
        """Return query results as geohashes.

        Args:
            precision (int): precision of the returned geohash

        Returns:
            HashesResponse
        """
        self.output("HASHES", precision)

        return HashesResponse(**(await self.exec()))

    async def asCount(self) -> CountResponse:
        """Return query results as count only.

        Args:

        Returns:
            CountResponse
        """
        self.output("COUNT")

        return CountResponse(**(await self.exec()))

    async def asIds(self) -> IdsResponse:
        """Return query results as object ids only.

        Args:

        Returns:
            IdsResponse
        """
        self.output("IDS")

        return IdsResponse(**(await self.exec()))

    async def asPoints(self) -> PointsResponse:
        """Return query results as points.

        Args:

        Returns:
            PointsResponse
        """
        self.output("POINTS")

        return PointsResponse(**(await self.exec()))

    def __compile_options(self) -> CommandArgs:
        """__compile_options.

        Args:

        Returns:
            CommandArgs
        """
        commands: list[CommandArg] = []

        # raises mypy: TypedDict key must be string literal
        # open PR: https://github.com/python/mypy/issues/7867
        for k in self._options:
            if isinstance(self._options[k], bool):  # type: ignore[literal-required]
                commands.append(k.upper())
            elif self._options[k] or self._options[k] == 0:  # type: ignore[literal-required]
                commands.extend([k.upper(), self._options[k]])  # type: ignore[literal-required]

        return commands

    def __compile_fence(self) -> CommandArgs:
        """__compile_fence

        Args:

        Returns:
            CommandArgs
        """
        return (
            [
                SubCommand.FENCE.value,
                *(
                    [SubCommand.DETECT.value, ",".join(self._detect)]
                    if self._detect
                    else []
                ),
                *(
                    [SubCommand.COMMANDS.value, ",".join(self._commands)]
                    if self._commands
                    else []
                ),
            ]
            if self._fence
            else []
        )

    def compile(self) -> Compiled:
        """compile

        Args:

        Returns:
            Compiled
        """
        compiled = [
            Command.WITHIN.value,
            [
                self._key,
                *(self.__compile_options()),
                *(self.compile_where()),
                *(self.compile_wherein()),
                *(self.__compile_fence()),
                *(self._output if self._output else []),
                *(self._query.get()),
            ],
        ]

        if self._hook:
            command, args = self._hook.compile()
            response: Compiled = [command, [*(args), *(flatten(compiled))]]  # type: ignore
            return response

        return compiled  # type: ignore

    async def activate(self) -> JSONResponse:
        """Activate is used in SetHook to activate a geo-fenced search.

        Args:

        Returns:
            JSONResponse
        """
        if self._hook:
            return JSONResponse(**(await self.client.command(*self.compile())))
        else:
            raise Pyle38NoHookToActivateError



import pytest

from pyle38 import Tile38
from pyle38.commands.executable import Compiled
from pyle38.commands.nearby import Format, Nearby

from .helper.random_data import random_string

key = random_string()
oid = random_string()
feature = {
    "type": "Feature",
    "geometry": {"type": "Point", "coordinates": [13.37, 52.25]},
    "properties": {"id": oid},
}

expected = {"id": oid, "object": feature}


@pytest.mark.parametrize(
    "fmt, expected",
    [
        (
            "OBJECTS",
            [
                "NEARBY",
                [
                    key,
                    "MATCH",
                    "*",
                    "NOFIELDS",
                    "SPARSE",
                    1,
                    "DISTANCE",
                    "CURSOR",
                    0,
                    "LIMIT",
                    10,
                    "WHERE",
                    "foo",
                    1,
                    1,
                    "WHERE",
                    "bar",
                    1,
                    1,
                    "WHEREIN",
                    "foo",
                    1,
                    1,
                    "WHEREIN",
                    "bar",
                    1,
                    1,
                    "FENCE",
                    "DETECT",
                    "enter,exit",
                    "COMMANDS",
                    "del,set",
                    "POINT",
                    1.0,
                    1.0,
                    100,
                ],
            ],
        )
    ],
    ids=["OBJECTS"],
)
@pytest.mark.asyncio
async def test_command_nearby_compile(
    tile38: Tile38, fmt: Format, expected: Compiled
) -> None:
    query = (
        Nearby(tile38.client, key)
        .match("*")
        .nofields()
        .sparse(1)
        .distance()
        .cursor(0)
        .limit(10)
        .where("foo", 1, 1)
        .where("bar", 1, 1)
        .wherein("foo", [1])
        .wherein("bar", [1])
        .fence()
        .detect(["enter", "exit"])
        .commands(["del", "set"])
        .point(1, 1, 100)
    )

    received = query.output(fmt).compile()

    assert expected == received


@pytest.mark.asyncio
async def test_command_nearby_point(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_nearby_where_point(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = (
        await tile38
        .nearby(key)
        .where("maxspeed", 120, 120)
        .point(52.250212, 13.370871)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .nearby(key)
        .where("maxspeed", 100, 120)
        .where("maxweight", 1000, 1000)
        .point(52.250212, 13.370871)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_nearby_wherein_point(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = (
        await tile38
        .nearby(key)
        .wherein("maxspeed", [120])
        .point(52.250212, 13.370871)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .nearby(key)
        .wherein("maxspeed", [100, 120])
        .wherein("maxweight", [1000])
        .point(52.250212, 13.370871)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_nearby_point_with_radius(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871, 10).asObjects()
    assert response.ok
    assert len(response.objects) == 0


@pytest.mark.asyncio
async def test_command_nearby_return_points(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asPoints()
    assert response.ok
    assert response.points[0].dict() == {
        "id": oid,
        "point": {"lat": 52.25, "lon": 13.37},
    }


@pytest.mark.asyncio
async def test_command_nearby_return_ids(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asIds()
    assert response.ok
    assert response.ids == [oid]


@pytest.mark.asyncio
async def test_command_nearby_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asCount()
    assert response.ok
    assert response.count == 1


@pytest.mark.asyncio
async def test_command_nearby_return_hashes(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asHashes(5)
    assert response.ok
    assert response.hashes[0].dict() == {"id": oid, "hash": "u3390"}


@pytest.mark.asyncio
async def test_command_nearby_return_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.nearby(key).point(52.250212, 13.370871).asBounds()
    assert response.ok
    assert response.bounds[0].dict() == {
        "id": oid,
        "bounds": {
            "ne": {"lat": 52.25, "lon": 13.37},
            "sw": {"lat": 52.25, "lon": 13.37},
        },
    }



import pytest

from pyle38 import Tile38

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_exists(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()

    await tile38.set(key, oid).point(1, 1).exec()

    received = await tile38.exists(key, oid)
    assert received.exists

    received = await tile38.exists(key, "does_not_exist")
    assert not received.exists



import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38IdNotFoundError, Tile38KeyNotFoundError

from .helper.random_data import random_latitude, random_longitude, random_string


@pytest.mark.asyncio
async def test_command_pdel(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    lat = random_latitude()
    lon = random_longitude()

    pdel_pattern = f"{oid[0:3]}*"

    response = await tile38.set(key, oid).point(lat, lon).exec()
    assert response.ok

    response = await tile38.get(key, oid).asObject()
    assert response.ok
    assert response.object["type"] == "Point"

    response = await tile38.pdel(key, pdel_pattern)
    assert response.ok

    # key drops because only id deleted
    with pytest.raises(Tile38KeyNotFoundError):
        await tile38.get(key, oid).asObject()

    # key stays because only one id is deleted
    await tile38.set(key, oid).point(lat, lon).exec()
    await tile38.set(key, random_string()).point(lat, lon).exec()

    await tile38.pdel(key, pdel_pattern)

    with pytest.raises(Tile38IdNotFoundError):
        await tile38.get(key, oid).asObject()



import abc
from typing import TypeVar

from ..client import CommandArgs, SubCommand

TWhereable = TypeVar("TWhereable", bound="Whereable")

Where = list[list[str | int]]
Wherein = list[list[str | int | float]]


class Whereable(abc.ABC):
    _where: Where
    _wherein: Wherein

    def compile_where(self) -> CommandArgs:
        """__compile_where.

        Args:

        Returns:
            CommandArgs
        """
        w = []

        if len(self._where) > 0:
            for i in self._where:
                w.extend(i)
            return w

        return w

    def compile_wherein(self) -> CommandArgs:
        """__compile_wherein.

        Args:

        Returns:
            CommandArgs
        """
        w = []

        if len(self._wherein) > 0:
            for i in self._wherein:
                w.extend(i)
            return w
        return []

    def wherein(
        self: TWhereable, field: str, values: list[int | str | float]
    ) -> TWhereable:
        """Filter the search by fields values containing input values

        Args:
            field (str): field name
            values (list): values to lookup in field values

        Returns:
            TWhereable
        """
        self._wherein.append([SubCommand.WHEREIN, field, len(values), *values])

        return self

    def where(self: TWhereable, field: str, minimum: int, maximum: int) -> TWhereable:
        """Filter the search by field

        Args:
            field (str): field name
            minimum (int): minimum value of field
            maximum (int): maximum value of field

        Returns:
            TWhereable
        """

        self._where.append([SubCommand.WHERE, field, minimum, maximum])

        return self

    def where_expr(self: TWhereable, expr: str) -> TWhereable:
        """Filter the search with an expression

        Args:
            expr (str): expression

        Returns:
            TWhereable

        Example:

        ```python
        await tile38.set('fleet', 'truck').fields({ 'driver': 'John'})
        await tile38.scan('fleet').where_expr("driver === 'John'")
        ```

        """

        self._where.append([SubCommand.WHERE, expr])

        return self



import pytest

from pyle38 import Tile38
from pyle38.commands.scan import Scan

from .helper.random_data import random_string

key = random_string()
oid = random_string()
feature = {
    "type": "Feature",
    "geometry": {"type": "Point", "coordinates": [13.37, 52.25]},
    "properties": {"id": oid},
}

expected = {"id": oid, "object": feature}


@pytest.mark.asyncio
async def test_command_scan_compile(tile38: Tile38) -> None:
    query = (
        Scan(tile38.client, key)
        .match("*")
        .nofields()
        .sparse(1)
        .asc()
        .cursor(0)
        .limit(10)
    )

    received = query.output("OBJECTS").compile()

    assert received == [
        "SCAN",
        [key, "MATCH", "*", "NOFIELDS", "SPARSE", 1, "ASC", "CURSOR", 0, "LIMIT", 10],
    ]

    query = (
        Scan(tile38.client, key)
        .match("*")
        .nofields()
        .sparse(1)
        .desc()
        .cursor(0)
        .limit(10)
    )

    received = query.output("OBJECTS").compile()

    assert received == [
        "SCAN",
        [key, "MATCH", "*", "NOFIELDS", "SPARSE", 1, "DESC", "CURSOR", 0, "LIMIT", 10],
    ]

    query = (
        Scan(tile38.client, key)
        .match("*")
        .nofields()
        .sparse(1)
        .desc()
        .cursor(0)
        .limit(10)
        .where("foo", 1, 1)
        .where("bar", 1, 1)
        .wherein("foo", [1])
        .wherein("bar", [1])
    )

    received = query.output("OBJECTS").compile()

    assert received == [
        "SCAN",
        [
            key,
            "MATCH",
            "*",
            "NOFIELDS",
            "SPARSE",
            1,
            "DESC",
            "CURSOR",
            0,
            "LIMIT",
            10,
            "WHERE",
            "foo",
            1,
            1,
            "WHERE",
            "bar",
            1,
            1,
            "WHEREIN",
            "foo",
            1,
            1,
            "WHEREIN",
            "bar",
            1,
            1,
        ],
    ]


@pytest.mark.asyncio
async def test_command_scan(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_scan_wherein(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = await tile38.scan(key).wherein("maxspeed", [120]).asObjects()
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .scan(key)
        .wherein("maxspeed", [100, 120])
        .wherein("maxweight", [1000])
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_scan_where(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = await tile38.scan(key).where("maxspeed", 120, 120).asObjects()
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .scan(key)
        .where("maxspeed", 100, 120)
        .where("maxweight", 1000, 1000)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_scan_return_points(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asPoints()
    assert response.ok
    assert response.points[0].dict() == {
        "id": oid,
        "point": {"lat": 52.25, "lon": 13.37},
    }


@pytest.mark.asyncio
async def test_command_scan_return_ids(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asIds()
    assert response.ok
    assert response.ids == [oid]


@pytest.mark.asyncio
async def test_command_scan_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asCount()
    assert response.ok
    assert response.count == 1


@pytest.mark.asyncio
async def test_command_scan_return_hashes(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asHashes(5)
    assert response.ok
    assert response.hashes[0].dict() == {"id": oid, "hash": "u3390"}


@pytest.mark.asyncio
async def test_command_scan_return_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.scan(key).asBounds()
    assert response.ok
    assert response.bounds[0].dict() == {
        "id": oid,
        "bounds": {
            "ne": {"lat": 52.25, "lon": 13.37},
            "sw": {"lat": 52.25, "lon": 13.37},
        },
    }



import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_stats(tile38: Tile38) -> None:
    response = await tile38.stats(["fleet"])
    assert response.ok
    assert response.stats == []

    await tile38.set("fleet", "truck1").point(1, 1).exec()
    await tile38.set("zones", "parking1").bounds(1, 1, 1, 1).exec()

    response = await tile38.stats(["fleet", "zones"])
    assert response.ok
    assert response.stats and len(response.stats) == 2



from __future__ import annotations

import json
from collections.abc import Sequence
from typing import Any, Literal

from typing_extensions import override

from ..client import Client, Command, SubCommand
from ..responses import Fields, JSONResponse
from .executable import Compiled, Executable


class Set(Executable):
    _key: str
    _id: str
    _ex: int | None = None
    _nx_or_xx: Literal["NX", "XX"] | None = None
    _fields: Fields | None
    _input: (
        Sequence[
            Literal["POINT", "OBJECT", "BOUNDS", "HASH", "STRING"] | str | float | int
        ]
        | None
    )

    def __init__(self, client: Client, key: str, oid: str) -> None:
        super().__init__(client)

        self.key(key).id(oid)
        self._fields = {}

    def key(self, value: str) -> Set:
        self._key = value

        return self

    def id(self, value: str) -> Set:
        self._id = value

        return self

    def fields(self, fields: Fields) -> Set:
        self._fields = fields

        return self

    def ex(self, seconds: int) -> Set:
        if seconds:
            self._ex = seconds

        return self

    def nx(self, flag: bool = True) -> Set:
        if flag:
            self._nx_or_xx = "NX"

        return self

    def xx(self, flag: bool = True) -> Set:
        if flag:
            self._nx_or_xx = "XX"

        return self

    def object(self, value: dict) -> Set:
        self._input = ["OBJECT", json.dumps(value)]

        return self

    def point(self, lat: float, lon: float, z: float | None = None) -> Set:
        """Define a point as input.

        Args:
            lat (float): latitude of input point
            lon (float): longitude of input point
            z (float): height of input point (optional)

        Returns:
            Set
        """

        if z:
            self._input = ["POINT", lat, lon, z]
            return self

        self._input = ["POINT", lat, lon]

        return self

    def bounds(
        self, min_lat: float, min_lon: float, max_lat: float, max_lon: float
    ) -> Set:
        self._input = ["BOUNDS", min_lat, min_lon, max_lat, max_lon]

        return self

    def hash(self, value: str) -> Set:
        self._input = ["HASH", value]

        return self

    def string(self, value: str) -> Set:
        self._input = ["STRING", value]

        return self

    @staticmethod
    def __unpack_fields(fields: Fields) -> list[Any]:
        command = []
        for k, v in fields.items():
            if isinstance(v, dict):
                command.extend([SubCommand.FIELD.value, k, json.dumps(v)])
                continue

            command.extend([SubCommand.FIELD.value, k, v])

        return command

    def compile(self) -> Compiled:
        return [
            Command.SET.value,
            [
                self._key,
                self._id,
                *(Set.__unpack_fields(self._fields) if self._fields else []),
                *([SubCommand.EX.value, self._ex] if self._ex else []),
                *([self._nx_or_xx] if self._nx_or_xx else []),
                *(self._input if self._input else []),
            ],
        ]  # type: ignore

    @override
    async def exec(self) -> JSONResponse:  # type: ignore[override]
        return JSONResponse(**(await self.client.command(*self.compile())))



import pytest

from pyle38.errors import Pyle38NoFollowerSetError
from pyle38.follower import Follower


@pytest.mark.asyncio
async def test_missing_follower_raises() -> None:
    with pytest.raises(Pyle38NoFollowerSetError):
        Follower("")



import pytest

from pyle38 import Tile38

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_bounds(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()

    response = await tile38.set(key, oid).point(1, 1).exec()
    assert response.ok

    expected = {
        "bounds": {
            "type": "Polygon",
            "coordinates": [
                [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]]
            ],
        },
    }
    received = await tile38.bounds(key)

    assert received.ok is True
    assert received.bounds.dict() == expected["bounds"]



from typing import Any

import pytest

from pyle38 import Tile38
from pyle38.commands.intersects import Intersects
from pyle38.errors import Pyle38BadObjectInputError
from pyle38.models import Feature, Polygon

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_intersects_compile(tile38: Tile38) -> None:
    key = random_string()

    expected = [
        "INTERSECTS",
        [
            key,
            "MATCH",
            "*",
            "BUFFER",
            10,
            "NOFIELDS",
            "SPARSE",
            1,
            "CLIP",
            "CURSOR",
            0,
            "LIMIT",
            10,
            "WHERE",
            "foo",
            1,
            1,
            "WHERE",
            "bar == 1",
            "WHEREIN",
            "foo",
            1,
            1,
            "WHEREIN",
            "bar",
            1,
            1,
            "FENCE",
            "DETECT",
            "enter,exit",
            "COMMANDS",
            "del,set",
            "CIRCLE",
            1.0,
            1.0,
            100,
        ],
    ]

    query = (
        Intersects(tile38.client, key)
        .match("*")
        .buffer(10)
        .nofields()
        .sparse(1)
        .clip()
        .cursor(0)
        .limit(10)
        .where("foo", 1, 1)
        .where_expr("bar == 1")
        .wherein("foo", [1])
        .wherein("bar", [1])
        .fence()
        .detect(["enter", "exit"])
        .commands(["del", "set"])
        .circle(1, 1, 100)
    )

    received = query.output("OBJECTS", None).compile()

    assert expected == received


key = random_string()
oid = random_string()
feature = {
    "type": "Feature",
    "geometry": {"type": "Point", "coordinates": [13.37, 52.25]},
    "properties": {"id": oid},
}
polygon: dict[str, Any] = {
    "type": "Polygon",
    "coordinates": [
        [
            [13.361263275146484, 52.24630137198303],
            [13.379974365234373, 52.24630137198303],
            [13.379974365234373, 52.256705331409506],
            [13.361263275146484, 52.256705331409506],
            [13.361263275146484, 52.24630137198303],
        ]
    ],
}

expected = {"id": oid, "object": feature}


@pytest.mark.asyncio
async def test_command_intersects_with_fields(tile38: Tile38) -> None:
    # field type str
    name = "Tom"
    # field type float
    weight = 111.1
    # field type int
    height = 190
    # field type dict
    info = {"height": height, "weight": weight, "name": name}

    response = (
        await tile38
        .set(key, oid)
        .fields({"info": info, "height": height, "weight": weight, "name": name})
        .object(feature)
        .exec()
    )
    assert response.ok

    response = (
        await tile38.intersects(key).bounds(52.24, 13.36, 52.256, 13.379).asObjects()
    )
    assert response.ok
    assert response.objects[0].dict() == {
        **expected,
        # in lexical order
        "fields": [height, info, name, weight],
    }


@pytest.mark.asyncio
async def test_command_intersects_circle(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_where_circle(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = (
        await tile38
        .intersects(key)
        .where("maxspeed", 120, 120)
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .intersects(key)
        .where("maxspeed", 100, 120)
        .where("maxweight", 1000, 1000)
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_intersects_wherein_circle(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = (
        await tile38
        .intersects(key)
        .wherein("maxspeed", [120])
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120, 1000]})

    response = (
        await tile38
        .intersects(key)
        .wherein("maxspeed", [100, 120])
        .wherein("maxweight", [1000])
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_intersects_where_expr_circle(tile38: Tile38) -> None:
    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .object(feature)
        .exec()
    )
    await (
        tile38
        .set(key, "truck1")
        .fields({"maxspeed": 100, "maxweight": 1000})
        .object(feature)
        .exec()
    )

    response = (
        await tile38
        .intersects(key)
        .where_expr("maxspeed >= 100 && maxspeed <= 120 && maxweight == 1000")
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_intersects_object(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    polygon: dict[str, Any] = {
        "type": "Polygon",
        "coordinates": [
            [
                [13.361263275146484, 52.24630137198303],
                [13.379974365234373, 52.24630137198303],
                [13.379974365234373, 52.256705331409506],
                [13.361263275146484, 52.256705331409506],
                [13.361263275146484, 52.24630137198303],
            ]
        ],
    }
    featurePolygon: dict[str, Any] = {
        "type": "Feature",
        "properties": {},
        "geometry": polygon,
    }

    response = await tile38.intersects(key).object(polygon).asObjects()

    assert response.ok
    assert response.objects[0].dict() == expected

    response = await tile38.intersects(key).object(Polygon(**polygon)).asObjects()

    assert response.ok
    assert response.objects[0].dict() == expected

    response = (
        await tile38.intersects(key).object(Feature(**featurePolygon)).asObjects()
    )

    assert response.ok
    assert response.objects[0].dict() == expected

    response = await tile38.intersects(key).object(featurePolygon).asObjects()

    assert response.ok
    assert response.objects[0].dict() == expected

    with pytest.raises(Pyle38BadObjectInputError):
        response = (
            await tile38
            .intersects(key)
            .object({
                "type": "Point",
                "coordinates": [1, 1],
            })
            .asObjects()
        )


@pytest.mark.asyncio
async def test_command_intersects_hash(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).hash("u3390").asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_quadkey(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).quadkey("120").asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_tile(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).tile(2200, 1348, 12).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_sector(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = (
        await tile38
        .intersects(key)
        .sector(52.25191, 13.37230, 1000, 180, 270)
        .asObjects()
    )
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = (
        await tile38.intersects(key).bounds(52.24, 13.36, 52.256, 13.379).asObjects()
    )
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_intersects_object_with_intersection(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()

    response = await tile38.set(key, oid).object(polygon).exec()
    assert response.ok

    intersecting_feature = {
        "type": "Polygon",
        "coordinates": [
            [
                [13.374652862548828, 52.24398904962714],
                [13.38357925415039, 52.24398904962714],
                [13.38357925415039, 52.24987472405909],
                [13.374652862548828, 52.24987472405909],
                [13.374652862548828, 52.24398904962714],
            ]
        ],
    }

    response = await tile38.intersects(key).object(intersecting_feature).asObjects()

    assert response.ok
    assert response.objects[0].dict() == {"id": oid, "object": polygon}


@pytest.mark.asyncio
async def test_command_intersects_get(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    oid2 = random_string()

    response = await tile38.set(key, oid).object(polygon).exec()
    assert response.ok

    intersecting_feature = {
        "type": "Polygon",
        "coordinates": [
            [
                [13.374652862548828, 52.24398904962714],
                [13.38357925415039, 52.24398904962714],
                [13.38357925415039, 52.24987472405909],
                [13.374652862548828, 52.24987472405909],
                [13.374652862548828, 52.24398904962714],
            ]
        ],
    }

    response = await tile38.set(key, oid2).object(intersecting_feature).exec()
    assert response.ok

    response = await tile38.intersects(key).get(key, oid2).asObjects()

    assert response.ok
    assert response.objects[0].dict() == {"id": oid, "object": polygon}


@pytest.mark.asyncio
async def test_command_intersects_return_points(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asPoints()
    assert response.ok
    assert response.points[0].dict() == {
        "id": oid,
        "point": {"lat": 52.25, "lon": 13.37},
    }


@pytest.mark.asyncio
async def test_command_intersects_return_ids(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asIds()
    assert response.ok
    assert response.ids == [oid]


@pytest.mark.asyncio
async def test_command_intersects_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asCount()
    assert response.ok
    assert response.count == 1


@pytest.mark.asyncio
async def test_command_intersects_return_hashes(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asHashes(5)
    assert response.ok
    assert response.hashes[0].dict() == {"id": oid, "hash": "u3390"}


@pytest.mark.asyncio
async def test_command_intersects_return_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.intersects(key).circle(52.25, 13.37, 100).asBounds()
    assert response.ok
    assert response.bounds[0].dict() == {
        "id": oid,
        "bounds": {
            "ne": {"lat": 52.25, "lon": 13.37},
            "sw": {"lat": 52.25, "lon": 13.37},
        },
    }


@pytest.mark.asyncio
async def test_command_intersects_buffer_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    search_area = {
        "type": "Feature",
        "properties": {},
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [13.37009847164154, 52.2498254610514],
                    [13.370516896247862, 52.2498254610514],
                    [13.370516896247862, 52.25017851139772],
                    [13.37009847164154, 52.25017851139772],
                    [13.37009847164154, 52.2498254610514],
                ]
            ],
        },
    }

    response = await tile38.intersects(key).object(search_area).asCount()
    assert response.ok
    assert response.count == 0

    response = await tile38.intersects(key).buffer(10).object(search_area).asCount()
    assert response.ok
    assert response.count == 1



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from ..client import Client, Command, SubCommand
from ..responses import BoundsNeSwResponse, HashResponse, ObjectResponse, PointResponse
from .executable import Compiled, Executable

Output = Sequence[Literal["HASH", "OBJECT", "POINT", "BOUNDS"] | int]

Formats = Literal["BOUNDS", "HASH", "OBJECT", "POINT"]


class Get(Executable):
    _key: str
    _id: str
    _withfields: Literal["WITHFIELDS"] | None = None
    _output: Output | None = None

    def __init__(self, client: Client, key: str, oid: str) -> None:
        super().__init__(client)

        self.key(key).id(oid)

    def key(self, key: str) -> Get:
        self._key = key

        return self

    def id(self, oid: str) -> Get:
        self._id = oid

        return self

    def withfields(self, flag: bool = True) -> Get:
        if flag:
            self._withfields = "WITHFIELDS"

        return self

    def output(self, fmt: Formats, precision: int | None = None) -> Get:
        if fmt == "OBJECT":
            self._output = None
        elif fmt == "HASH" and precision:
            self._output = [fmt, precision]
        elif fmt == "BOUNDS" or fmt == "POINT":
            self._output = [fmt]

        return self

    async def asObject(self) -> ObjectResponse:
        self.output("OBJECT")

        return ObjectResponse(**(await self.exec()))

    async def asStringObject(self) -> ObjectResponse[str]:
        self.output("OBJECT")

        return ObjectResponse[str](**(await self.exec()))

    async def asBounds(self) -> BoundsNeSwResponse:
        self.output("BOUNDS")

        return BoundsNeSwResponse(**(await self.exec()))

    async def asPoint(self) -> PointResponse:
        self.output("POINT")

        return PointResponse(**(await self.exec()))

    async def asHash(self, precision: int) -> HashResponse:
        self.output("HASH", precision)

        return HashResponse(**(await self.exec()))

    def compile(self) -> Compiled:
        return [
            Command.GET.value,
            [
                self._key,
                self._id,
                *([SubCommand.WITHFIELDS.value] if self._withfields else []),
                *(self._output if self._output else []),
            ],
        ]  # type: ignore



from __future__ import annotations

from pyle38.errors import Tile38FieldNotFoundError
from pyle38.responses import JSONGetResponse

from ..client import Client, Command
from .executable import Compiled


class FGet:
    """Get the value for a field of an id.

    Example::

        key = fleet
        id = truck1
        fields = {"driver":"John"}
        await tile38.set(key, id).point(52.25, 13.37).exec()
        await tile38.fget(key, id, "driver")
    """

    _client: Client
    _key: str
    _id: str
    _field: str

    def __init__(self, client: Client, key: str, oid: str, field: str) -> None:
        self.client = client
        self.key(key).id(oid).field(field)

    def key(self, key: str) -> FGet:
        """Set key of id to get field for

        Args:
        key (str): key of a collection

        Returns:
            Self
        """
        self._key = key

        return self

    def id(self, oid: str) -> FGet:
        """Set id to get field for

        Args:
        id (str): id of an object

        Returns:
            Self
        """
        self._id = oid

        return self

    def field(self, field: str) -> FGet:
        """Set field to get

        Args:
        field (str): field name of an object

        Returns:
            Self
        """
        self._field = field

        return self

    def compile(self) -> Compiled:
        return [
            Command.FGET.value,
            [
                self._key,
                self._id,
                self._field,
            ],
        ]  # type: ignore

    async def exec(self) -> JSONGetResponse:
        """Execute the FGet command with input parameters.

        Zero Value fields that return as {"ok":true,"value":0} from Tile38
        are fields that are not set with Tile38, hence we know they do not exist
        and can raise a Tile38FieldNotFoundError.

        Args:
        field (str): field name of an object

        Returns:
            Self
        """
        resp = await self.client.command(*self.compile())

        if resp["ok"] and resp["value"] == 0:
            msg = "field not found"
            raise Tile38FieldNotFoundError(msg)

        return JSONGetResponse(**resp)



import abc
from typing import Any

from ..client import Client, Command, CommandArgs

Compiled = tuple[Command, CommandArgs]


class Executable(abc.ABC):
    """Executable is an abstract base class for Tile38-style command execution."""

    client: Client

    def __init__(self, client: Client) -> None:
        """__init__.

        Args:
            client (Client): The client instance used to communicate with the backend.

        Returns:
            None
        """
        self.client = client

    def compile(self) -> Compiled:
        """compile.

        This method should be overridden by subclasses to provide the command and its arguments.

        Returns:
            Compiled: A tuple containing the Command and its associated arguments.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("Not implemented")

    async def exec(self) -> dict[Any, Any]:
        """exec.

        Executes the compiled command asynchronously using the associated client.

        Returns:
            dict[Any, Any]: The response from the client command execution.
        """
        return await self.client.command(*self.compile())



import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_server_extended(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    response = await tile38.server_extended()
    assert response.ok
    assert response.stats.tile38_connected_slaves == 1

    response = await tile38.follower().server_extended()
    assert response.ok
    assert response.stats.tile38_connected_slaves == 0



import pytest

from pyle38 import Tile38
from pyle38.commands.search import Search

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_search_compile(tile38: Tile38) -> None:
    key = random_string()

    query = Search(tile38.client, key).match("*").asc().cursor(0).limit(10)

    received = query.output("OBJECTS").compile()

    assert received == ["SEARCH", [key, "MATCH", "*", "ASC", "CURSOR", 0, "LIMIT", 10]]

    query = Search(tile38.client, key).match("*").desc().cursor(0).limit(10)

    received = query.output("OBJECTS").compile()

    assert received == ["SEARCH", [key, "MATCH", "*", "DESC", "CURSOR", 0, "LIMIT", 10]]

    query = (
        Search(tile38.client, key)
        .match("*")
        .where("foo", 1, 1)
        .wherein("foo", [1])
        .desc()
        .cursor(0)
        .limit(10)
    )

    received = query.output("OBJECTS").compile()

    assert received == [
        "SEARCH",
        [
            key,
            "MATCH",
            "*",
            "DESC",
            "CURSOR",
            0,
            "LIMIT",
            10,
            "WHERE",
            "foo",
            1,
            1,
            "WHEREIN",
            "foo",
            1,
            1,
        ],
    ]


@pytest.mark.asyncio
async def test_command_search_returns_stringobjects(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    string = random_string()

    pattern = string[:1] + "*"

    response = await tile38.set(key, oid).string(string).exec()
    assert response.ok

    response = await tile38.search(key).match(pattern).asStringObjects()
    assert response.ok
    assert response.objects[0].dict()["object"] == string


@pytest.mark.asyncio
async def test_command_search_returns_where_stringobjects(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    string = random_string()
    pattern = string[:1] + "*"

    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .string(string)
        .exec()
    )
    await (
        tile38
        .set(key, random_string())
        .fields({"maxspeed": 100, "maxweight": 1000})
        .string(random_string())
        .exec()
    )

    response = (
        await tile38
        .search(key)
        .where("maxspeed", 120, 120)
        .match(pattern)
        .asStringObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(
        {"id": oid, "object": string}, **{"fields": [120, 1000]}
    )

    response = (
        await tile38
        .search(key)
        .where("maxspeed", 100, 120)
        .where("maxweight", 1000, 1000)
        .match(pattern)
        .asStringObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(
        {"id": oid, "object": string}, **{"fields": [120, 1000]}
    )


@pytest.mark.asyncio
async def test_command_search_returns_wherein_stringobjects(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    string = random_string()
    pattern = string[:1] + "*"

    await (
        tile38
        .set(key, oid)
        .fields({"maxspeed": 120, "maxweight": 1000})
        .string(string)
        .exec()
    )
    await (
        tile38
        .set(key, random_string())
        .fields({"maxspeed": 100, "maxweight": 1000})
        .string(random_string())
        .exec()
    )

    response = (
        await tile38
        .search(key)
        .wherein("maxspeed", [120])
        .match(pattern)
        .asStringObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(
        {"id": oid, "object": string}, **{"fields": [120, 1000]}
    )

    response = (
        await tile38
        .search(key)
        .wherein("maxspeed", [100, 120])
        .wherein("maxweight", [1000])
        .match(pattern)
        .asStringObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(
        {"id": oid, "object": string}, **{"fields": [120, 1000]}
    )


@pytest.mark.asyncio
async def test_command_search_returns_ids(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    string = random_string()
    pattern = string[:1] + "*"

    response = await tile38.set(key, oid).string(string).exec()
    assert response.ok

    response = await tile38.search(key).match(pattern).asIds()
    assert response.ok
    assert response.ids == [oid]


@pytest.mark.asyncio
async def test_command_search_returns_count(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    string = random_string()
    pattern = string[:1] + "*"

    response = await tile38.set(key, oid).string(string).exec()
    assert response.ok

    response = await tile38.search(key).match(pattern).asCount()
    assert response.ok
    assert response.count == 1



import pytest

from pyle38 import Tile38

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_fexists(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    field = random_string()
    value = random_string()

    await tile38.set(key, oid).fields({field: value}).point(1, 1).exec()

    received = await tile38.fexists(key, oid, field)
    assert received.exists

    received = await tile38.fexists(key, oid, "does_not_exist")
    assert not received.exists



import pytest

from pyle38 import Tile38

from .helper.random_data import random_point_feature, random_string

key = random_string()
oid = random_string()
fields = {"speed": 100, "state": 1}
feature = random_point_feature(
    {"type": "Point", "coordinates": [13.37, 52.25]}, {"id": oid}
)


@pytest.mark.asyncio
async def test_option_fields_in_objects(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).fields(fields).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asObjects()
    assert response.ok
    assert response.fields == ["speed", "state"]
    assert response.objects[0].dict() == {
        "id": oid,
        "object": feature,
        "fields": [100, 1],
    }


@pytest.mark.asyncio
async def test_option_fields_in_hashes(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).fields(fields).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asHashes(5)
    assert response.ok
    assert response.fields == ["speed", "state"]
    assert response.hashes[0].dict() == {"id": oid, "hash": "u3390", "fields": [100, 1]}


@pytest.mark.asyncio
async def test_option_fields_in_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).fields(fields).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asBounds()
    assert response.ok
    assert response.fields == ["speed", "state"]
    assert response.bounds[0].dict() == {
        "id": oid,
        "bounds": {
            "ne": {"lat": 52.25, "lon": 13.37},
            "sw": {"lat": 52.25, "lon": 13.37},
        },
        "fields": [100, 1],
    }


@pytest.mark.asyncio
async def test_option_fields_in_points(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).fields(fields).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asPoints()
    assert response.ok
    assert response.fields == ["speed", "state"]
    assert response.points[0].dict() == {
        "id": oid,
        "point": {"lat": 52.25, "lon": 13.37},
        "fields": [100, 1],
    }



import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_ping(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    response = await tile38.ping()
    assert response.ok
    assert response.ping == "pong"

    response = await tile38.follower().ping()
    assert response.ok
    assert response.ping == "pong"



import json
from collections.abc import Sequence
from typing import Any, Literal, TypedDict

from pydantic import BaseModel


class Options(TypedDict, total=False):
    cursor: int | None
    limit: int | None
    nofields: int | None
    match: str | None
    sparse: int | None
    clip: bool | None
    distance: bool | None
    asc: bool | None
    desc: bool | None
    buffer: int | None


class CircleQuery(BaseModel):
    command: Literal["CIRCLE"] = "CIRCLE"
    lat: float
    lon: float
    radius: float

    def get(self) -> Sequence[str | int | float]:
        return [self.command, self.lat, self.lon, self.radius]


class PointQuery(BaseModel):
    command: Literal["POINT"] = "POINT"
    lat: float
    lon: float
    radius: float | int | None = None

    def get(self) -> Sequence[str | int | float]:
        if self.radius:
            return [self.command, self.lat, self.lon, self.radius]

        return [self.command, self.lat, self.lon]


class BoundsQuery(BaseModel):
    command: Literal["BOUNDS"] = "BOUNDS"
    minlat: float
    minlon: float
    maxlat: float
    maxlon: float

    def get(self) -> Sequence[str | int | float]:
        return [self.command, self.minlat, self.minlon, self.maxlat, self.maxlon]


class HashQuery(BaseModel):
    command: Literal["HASH"] = "HASH"
    geohash: str

    def get(self) -> Sequence[str]:
        return [self.command, self.geohash]


class QuadkeyQuery(BaseModel):
    command: Literal["QUADKEY"] = "QUADKEY"
    quadkey: str

    def get(self) -> Sequence[str]:
        return [self.command, self.quadkey]


class TileQuery(BaseModel):
    command: Literal["TILE"] = "TILE"
    x: int
    y: int
    z: int

    def get(self) -> Sequence[str | int]:
        return [self.command, self.x, self.y, self.z]


class SectorQuery(BaseModel):
    command: Literal["SECTOR"] = "SECTOR"
    lat: float
    lon: float
    radius: float
    bearing1: float
    bearing2: float

    def get(self) -> Sequence[str | float]:
        return [
            self.command,
            self.lat,
            self.lon,
            self.radius,
            self.bearing1,
            self.bearing2,
        ]


Coordinate = tuple[float, float]


class Polygon(BaseModel):
    type: Literal["Polygon"] = "Polygon"
    coordinates: list[list[Coordinate]]


class Feature(BaseModel):
    type: Literal["Feature"] = "Feature"
    geometry: Polygon
    properties: dict[Any, Any]


class ObjectQuery(BaseModel):
    command: Literal["OBJECT"] = "OBJECT"
    object: Polygon | Feature

    def get(self) -> Sequence[str]:
        return [self.command, json.dumps(self.object.model_dump())]


class GetQuery(BaseModel):
    command: Literal["GET"] = "GET"
    key: str
    id: str

    def get(self) -> Sequence[str]:
        return [self.command, self.key, self.id]



import pytest

from pyle38 import Tile38
from pyle38.commands.executable import Compiled
from pyle38.commands.get import Formats, Get

from .helper.random_data import random_point_feature, random_string

key = random_string()
oid = random_string()
obj = random_point_feature()


@pytest.mark.parametrize(
    "fmt, precision, expected",
    [
        ("POINT", None, ["GET", [key, oid, "WITHFIELDS", "POINT"]]),
        ("BOUNDS", None, ["GET", [key, oid, "WITHFIELDS", "BOUNDS"]]),
        ("OBJECT", None, ["GET", [key, oid, "WITHFIELDS"]]),
        ("HASH", 5, ["GET", [key, oid, "WITHFIELDS", "HASH", 5]]),
    ],
    ids=["point", "bounds", "object", "hash"],
)
@pytest.mark.asyncio
async def test_command_get_compile(
    fmt: Formats, precision: int | None, expected: Compiled, tile38: Tile38
) -> None:
    query = Get(tile38.client, key, oid).withfields()
    received = query.output(fmt, precision).compile()
    assert expected == received


@pytest.mark.asyncio
async def test_command_get_query(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    obj = {
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": [1, 1]},
        "properties": {},
    }

    await tile38.set(key, oid).object(obj).exec()
    await tile38.set(key, f"{oid}:driver").string("John").exec()

    obj_response = await tile38.get(key, oid).asObject()
    assert obj == obj_response.object

    expected_point = {"lat": 1, "lon": 1}
    point_response = await tile38.get(key, oid).asPoint()
    assert point_response.point.dict() == expected_point

    expected_bounds = {"ne": {"lat": 1, "lon": 1}, "sw": {"lat": 1, "lon": 1}}
    bounds_response = await tile38.get(key, oid).asBounds()
    assert bounds_response.bounds.dict() == expected_bounds

    expected_hash = "s00twy0"
    hash_response = await tile38.get(key, oid).asHash(7)
    assert expected_hash == hash_response.hash

    expected_string = "John"
    str_obj_response = await tile38.get(key, f"{oid}:driver").asStringObject()
    assert str_obj_response.object == expected_string



import pytest

from pyle38 import Tile38


@pytest.mark.asyncio
async def test_command_info(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    leaderResponse = await tile38.info()
    assert leaderResponse.ok

    followerResponse = await tile38.follower().info()
    assert followerResponse.ok



import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38Error

from .helper.random_data import random_latitude, random_longitude, random_string


@pytest.mark.asyncio
async def test_command_readonly(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    lat = random_latitude()
    lon = random_longitude()

    response = await tile38.set(key, oid).point(lat, lon).exec()
    assert response.ok

    response = await tile38.readonly()
    assert response.ok

    with pytest.raises(Tile38Error):
        await tile38.set(key, oid).point(lat, lon).exec()

    response = await tile38.readonly(value=False)
    assert response.ok

    response = await tile38.server_extended()
    assert response.ok
    assert response.stats.tile38_read_only is False

    response = await tile38.set(key, oid).point(lat, lon).exec()
    assert response.ok



import pytest

from pyle38 import Tile38
from pyle38.commands.fset import FSet

from .helper.random_data import random_integer, random_point_feature, random_string


@pytest.mark.asyncio
async def test_command_fset_compile(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = random_integer(1, 10)
    fields = {fkey: fvalue}

    received = FSet(tile38.client, key, oid, fields).compile()
    expected = ["FSET", [key, oid, fkey, fvalue]]

    assert received == expected


@pytest.mark.asyncio
async def test_command_fset(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = random_integer(1, 10)
    fields = {fkey: fvalue}
    obj = random_point_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.fset(key, oid, fields).xx().exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields


@pytest.mark.asyncio
async def test_command_fset_expr(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = {random_string(): random_integer(1, 10)}
    fields = {fkey: fvalue}
    obj = random_point_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.fset(key, oid, fields).xx().exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields


@pytest.mark.asyncio
async def test_command_fset_zero_deletes_field(tile38: Tile38) -> None:
    """
    ZeroValue fields are treated as non-existent in the tile38 field value context.
    """
    key = random_string()
    oid = random_string()
    fkey = random_string()
    fvalue = {random_string(): random_integer(1, 10)}
    fields = {fkey: fvalue}
    obj = random_point_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    response = await tile38.fset(key, oid, fields).xx().exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields

    response = await tile38.fset(key, oid, {fkey: 0}).xx().exec()
    assert response.ok

    response = await tile38.fexists(key, oid, fkey)
    response.exists = False



from redis.exceptions import ConnectionError, RedisError, TimeoutError  # noqa: A004

Pyle38Error = RedisError
Pyle38ConnectionError = ConnectionError
Pyle38TimeoutError = TimeoutError


class Tile38Error(Exception):
    pass


class Tile38IdNotFoundError(Exception):
    pass


class Tile38KeyNotFoundError(Exception):
    pass


class Tile38FieldNotFoundError(Exception):
    pass


class Tile38NotCaughtUpError(Exception):
    pass


class Tile38PathNotFoundError(Exception):
    pass


class Pyle38BadObjectInputError(Exception):
    def __init__(self) -> None:
        return super().__init__(
            "Input object is type dict but neither Polygon nor Feature type"
        )


class Pyle38NoFollowerSetError(Exception):
    def __init__(self) -> None:
        return super().__init__("No Tile38 follower URI set")


class Pyle38NoLeaderSetError(Exception):
    def __init__(self) -> None:
        return super().__init__("No Tile38 leader URI set")


class Pyle38NoHookToActivateError(Exception):
    def __init__(self) -> None:
        return super().__init__("No hook to activate")



import os

import pytest

from pyle38 import Tile38
from pyle38.client import (
    Client,
)
from pyle38.client_options import WithRetryExponentialBackoff, WithRetryOnError
from pyle38.errors import (
    Pyle38ConnectionError,
    Pyle38NoFollowerSetError,
    Pyle38NoLeaderSetError,
    Pyle38TimeoutError,
    Tile38Error,
    Tile38IdNotFoundError,
    Tile38KeyNotFoundError,
)


@pytest.mark.asyncio
async def test_client_options() -> None:
    url = os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851"
    client_options = [
        WithRetryExponentialBackoff(10),
        WithRetryOnError(Pyle38TimeoutError, Pyle38ConnectionError),
    ]
    client = Client(url, client_options)

    assert client.url == url

    opts = client.client_options()
    assert opts["retry"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    assert type(opts["retry_on_error"]) is list  # pyright: ignore[reportTypedDictNotRequiredAccess]
    assert opts["retry_on_error"] == [Pyle38TimeoutError, Pyle38ConnectionError]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    response = await client.command("SET", ["fleet", "truck", "POINT", 1, 1])  # type: ignore[arg-type]
    assert response["ok"]

    response = await client.command("GET", ["fleet", "truck"])
    assert response["ok"]

    await client.quit()


@pytest.mark.asyncio
async def test_client() -> None:
    client = Client(os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851")
    response = await client.command("SET", ["fleet", "truck", "POINT", 1, 1])  # type: ignore[arg-type]
    assert response["ok"]

    response = await client.command("GET", ["fleet", "truck"])
    assert response["ok"]

    await client.quit()


@pytest.mark.asyncio
async def test_client_exceptions() -> None:
    client = Client(os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851")
    with pytest.raises(Tile38Error):
        await client.command("BLA")
    with pytest.raises(Tile38IdNotFoundError):
        await client.command("GET", ["fleet", "Scooter"])

    with pytest.raises(Tile38KeyNotFoundError):
        await client.command("GET", ["vehicles", "Scooter"])

    await client.quit()


@pytest.mark.asyncio
async def test_client_empty_uri() -> None:
    with pytest.raises(Pyle38NoLeaderSetError):
        client = Tile38("")
        await client.get("BLA", "BLA").asObject()


@pytest.mark.asyncio
async def test_no_follower_set() -> None:
    with pytest.raises(Pyle38NoFollowerSetError):
        client = Tile38(os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851")
        await client.follower().get("BLA", "BLA").asObject()


@pytest.mark.asyncio
async def test_client_quit() -> None:
    client = Client(os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851")
    response = await client.quit()

    assert response == "OK"


@pytest.mark.asyncio
async def test_client_handles_disconnection() -> None:
    client = Client(os.getenv("TILE38_LEADER_URI") or "redis://localhost:9851")

    response = await client.command("SET", ["fleet", "truck", "POINT", 1, 1])  # type: ignore[arg-type]
    assert response["ok"]

    quit_response = await client.quit()
    assert quit_response == "OK"

    response = await client.command("SET", ["fleet", "truck", "POINT", 1, 1])  # type: ignore[arg-type]
    assert response["ok"]

    quit_response = await client.quit()
    assert quit_response == "OK"



import json

import pytest

from pyle38 import Tile38
from pyle38.client import Client
from pyle38.commands.executable import Compiled
from pyle38.commands.set import Set

from .helper.random_data import (
    random_float,
    random_integer,
    random_polygon_feature,
    random_position,
    random_string,
)

client = Client("test")

key = random_string()
oid = random_string()
obj: dict = random_polygon_feature()
fields = {"speed": 100, "state": 1}


@pytest.mark.parametrize(
    "expected, received",
    [
        (
            ["SET", [key, oid, "NX", "POINT", 1, 1]],
            Set(client, key, oid).nx().point(1, 1).compile(),
        ),
        (
            ["SET", [key, oid, "NX", "POINT", 1, 1, 1]],
            Set(client, key, oid).nx().point(1, 1, 1).compile(),
        ),
        (
            ["SET", [key, oid, "EX", 5, "BOUNDS", 1, 1, 1, 1]],
            Set(client, key, oid).ex(5).bounds(1, 1, 1, 1).compile(),
        ),
        (
            ["SET", [key, oid, "XX", "OBJECT", json.dumps(obj)]],
            Set(client, key, oid).xx().object(obj).compile(),
        ),
        (
            ["SET", [key, oid, "HASH", "u33d"]],
            Set(client, key, oid).hash("u33d").compile(),
        ),
        (
            ["SET", [key, oid, "STRING", oid]],
            Set(client, key, oid).string(oid).compile(),
        ),
        (
            [
                "SET",
                [key, oid, "FIELD", "speed", 100, "FIELD", "state", 1, "POINT", 1, 1],
            ],
            Set(client, key, oid).fields(fields).point(1, 1).compile(),
        ),
    ],
    ids=["point", "point_z", "bounds", "object", "hash", "string", "with fields"],
)
@pytest.mark.asyncio
async def test_command_set_compile(expected: Compiled, received: Compiled) -> None:
    assert expected == received


@pytest.mark.asyncio
async def test_command_set_get(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()
    obj = random_polygon_feature()

    response = await tile38.set(key, oid).object(obj).exec()
    assert response.ok

    expected = obj
    received = await tile38.get(key, oid).asObject()

    assert received.ok
    assert expected == received.object


@pytest.mark.asyncio
async def test_command_set_get_point_z(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()

    response = await tile38.set(key, oid).point(1, 1, 1).exec()
    assert response.ok

    received = await tile38.get(key, oid).asPoint()

    assert received.ok
    assert received.point.lat == 1.0
    assert received.point.lon == 1.0
    assert received.point.z == 1.0

    response = await tile38.set(key, oid).point(1, 1).exec()
    assert response.ok

    received = await tile38.get(key, oid).asPoint()

    assert received.ok
    assert received.point.lat == 1.0
    assert received.point.lon == 1.0
    assert not received.point.z


@pytest.mark.parametrize(
    "key, value",
    [
        (
            "integer",
            random_integer(),
        ),
        (
            "float",
            random_float(),
        ),
        (
            "string",
            random_string(),
        ),
        (
            "dictionary",
            {
                "foo": random_string(),
                "bar": random_float(),
            },
        ),
    ],
    ids=["integer", "float", "string", "dictionary"],
)
@pytest.mark.asyncio
async def test_command_set_get_with_fields(
    tile38: Tile38, key: str, value: str | float | int | dict
) -> None:
    key = random_string()
    oid = random_string()
    [lng, lat] = random_position()

    fields = {key: value}

    response = await tile38.set(key, oid).fields(fields).point(lat, lng).exec()
    assert response.ok

    response = await tile38.get(key, oid).withfields().asObject()
    assert response.ok
    assert response.fields == fields



import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38NotCaughtUpError
from pyle38.parse_response import parse_response


@pytest.mark.asyncio
async def test_command_healthz(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    response = await tile38.healthz()
    assert response.ok

    response = await tile38.follower().healthz()
    assert response.ok


@pytest.mark.asyncio
async def test_parse_command_raises() -> None:
    with pytest.raises(Tile38NotCaughtUpError):
        parse_response('{"ok":false,"err":"not caught up","elapsed":"1ms"}')



from collections.abc import Callable
from typing import Literal

from .client import Client, Command, SubCommand
from .client_options import ClientOptions
from .commands.fget import FGet
from .commands.get import Get
from .commands.intersects import Intersects
from .commands.nearby import Nearby
from .commands.scan import Scan
from .commands.search import Search
from .commands.within import Within
from .errors import Pyle38NoFollowerSetError
from .responses import (
    BoundsResponse,
    ChansResponse,
    ConfigGetResponse,
    ConfigKeys,
    ExistsResponse,
    HooksResponse,
    InfoFollowerResponse,
    JSONGetResponse,
    JSONResponse,
    KeysResponse,
    PingResponse,
    ServerStatsExtendedResponse,
    ServerStatsResponseFollower,
    StatsResponse,
)


class Follower:
    """
    A class to interact with a Tile38 follower instance using various Tile38 commands.

    Attributes:
        client (Client): An instance of the Client class for connecting to Tile38.
    """

    client: Client

    def __init__(self, url: str, opts: list[Callable[..., ClientOptions]] = []) -> None:
        """Initialize the Follower client.

        Args:
            url (str): The URL for connecting to the Tile38 follower server.
            opts (List[Callable[..., ClientOptions]], optional): A list of callables that modify client options.

        Raises:
            Tile38Error: If the follower URL is not provided.
        """
        if not url:
            raise Pyle38NoFollowerSetError

        self.client = Client(url, opts)

    async def aofshrink(self) -> JSONResponse:
        """Shrink the append only file on"""
        return JSONResponse(**(await self.client.command(Command.AOFSHRINK)))

    async def exists(self, key: str, oid: str) -> ExistsResponse:
        """Check if an object exists in a collection.

        Tile38 Command:
            EXISTS key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            ExistsResponse: Response indicating if the object exists.
        """
        return ExistsResponse(**(await self.client.command(Command.EXISTS, [key, oid])))

    async def fexists(self, key: str, oid: str, field: str) -> ExistsResponse:
        """Check if a field exists within an object.

        Tile38 Command:
            FEXISTS key id field

        Args:
            key (str): The collection key.
            id (str): The object ID.
            field (str): The field name.

        Returns:
            ExistsResponse: Response indicating if the field exists.
        """
        return ExistsResponse(
            **(await self.client.command(Command.FEXISTS, [key, oid, field]))
        )

    async def bounds(self, key: str) -> BoundsResponse:
        """Get the bounds of a collection.

        Tile38 Command:
            BOUNDS key

        Args:
            key (str): The collection key.

        Returns:
            BoundsResponse: The bounding box of the specified collection.
        """
        return BoundsResponse(**(await self.client.command(Command.BOUNDS, [key])))

    async def chans(self, pattern: str = "*") -> ChansResponse:
        """List active channels.

        Tile38 Command:
            CHANS pattern

        Args:
            pattern (str, optional): The pattern to match channel names.

        Returns:
            ChansResponse: A list of active channels.
        """
        return ChansResponse(**(await self.client.command(Command.CHANS, [pattern])))

    async def config_get(self, name: ConfigKeys) -> ConfigGetResponse:
        """Get a configuration value.

        Tile38 Command:
            CONFIG GET name

        Args:
            name (ConfigKeys): The configuration key.

        Returns:
            ConfigGetResponse: The value of the requested configuration key.
        """
        return ConfigGetResponse(
            **(await self.client.command(Command.CONFIG, [Command.GET, name]))
        )

    async def config_set(
        self, name: ConfigKeys, value: str | int | float
    ) -> JSONResponse:
        """Set a configuration value.

        Tile38 Command:
            CONFIG SET name value

        Args:
            name (ConfigKeys): The configuration key.
            value (Union[str, int, float]): The value to set.

        Returns:
            JSONResponse: Confirmation of the configuration update.
        """
        return JSONResponse(
            **(await self.client.command(Command.CONFIG, [Command.SET, name, value]))
        )

    async def config_rewrite(self) -> JSONResponse:
        """Rewrite the configuration file.

        Tile38 Command:
            CONFIG REWRITE

        Returns:
            JSONResponse: Confirmation of the rewrite.
        """
        return JSONResponse(
            **(await self.client.command(Command.CONFIG, [SubCommand.REWRITE]))
        )

    async def gc(self) -> JSONResponse:
        """Trigger garbage collection.

        Tile38 Command:
            GC

        Returns:
            JSONResponse: Confirmation of garbage collection.
        """
        return JSONResponse(**(await self.client.command(Command.GC)))

    def get(self, key: str, oid: str) -> Get:
        """Get a specific object.

        Tile38 Command:
            GET key id

        Args:
            key (str): The collection key.
            id (str): The object ID.

        Returns:
            Get: An object that allows retrieving data from Tile38.
        """
        return Get(self.client, key, oid)

    def fget(self, key: str, oid: str, field: str) -> FGet:
        """Get a specific field on an object.

        Tile38 Command:
            FGET key id field

        Args:
            key (str): The collection key.
            id (str): The object ID.
            field (str): The field to look up.

        Returns:
            FGet: An object that allows retrieving field data from Tile38.
        """
        return FGet(self.client, key, oid, field)

    async def hooks(self, pattern: str = "*") -> HooksResponse:
        """List active hooks.

        Tile38 Command:
            HOOKS pattern

        Args:
            pattern (str, optional): The pattern to match hook names.

        Returns:
            HooksResponse: A list of active hooks.
        """
        return HooksResponse(**(await self.client.command(Command.HOOKS, [pattern])))

    async def healthz(self) -> JSONResponse:
        """Check the health of the Tile38 server.

        Tile38 Command:
            HEALTHZ

        Returns:
            JSONResponse: Server health status.
        """
        return JSONResponse(**(await self.client.command(Command.HEALTHZ)))

    async def info(self) -> InfoFollowerResponse:
        """Get server information.

        Tile38 Command:
            INFO

        Returns:
            InfoFollowerResponse: Server information and statistics.
        """
        return InfoFollowerResponse(**(await self.client.command(Command.INFO)))

    def intersects(self, key: str) -> Intersects:
        """Query objects that intersect a specified area.

        Tile38 Command:
            INTERSECTS

        Args:
            key (str): The collection key.

        Returns:
            Intersects: An object to perform intersection queries.
        """
        return Intersects(self.client, key)

    async def keys(self, pattern: str = "*") -> KeysResponse:
        """List keys matching a pattern.

        Tile38 Command:
            KEYS pattern

        Args:
            pattern (str, optional): The pattern to match keys.

        Returns:
            KeysResponse: A list of matching keys.
        """
        return KeysResponse(**(await self.client.command(Command.KEYS, [pattern])))

    async def jget(
        self,
        key: str,
        oid: str,
        path: str | None = None,
        mode: Literal["RAW"] | None = None,
    ) -> JSONGetResponse:
        """Get a JSON field.

        Tile38 Command:
            JGET key id [path] [RAW]

        Args:
            key (str): The collection key.
            id (str): The object ID.
            path (Optional[str], optional): The JSON path.
            mode (Optional[Literal["RAW"]], optional): RAW mode for unparsed output.

        Returns:
            JSONGetResponse: The requested JSON data.
        """
        return JSONGetResponse(
            **(
                await self.client.command(
                    Command.JGET,
                    [key, oid, *([path] if path else []), *([mode] if mode else [])],
                )
            )
        )

    def nearby(self, key: str) -> Nearby:
        """Find objects near a specified location.

        Tile38 Command:
            NEARBY

        Args:
            key (str): The collection key.

        Returns:
            Nearby: An object to perform nearby queries.
        """
        return Nearby(self.client, key)

    async def ping(self) -> PingResponse:
        """Ping the Tile38 server.

        Tile38 Command:
            PING

        Returns:
            PingResponse: Ping response from the server.
        """
        return PingResponse(**(await self.client.command(Command.PING)))

    def scan(self, key: str) -> Scan:
        """Scan a collection.

        Tile38 Command:
            SCAN

        Args:
            key (str): The collection key.

        Returns:
            Scan: An object to perform scan queries.
        """
        return Scan(self.client, key)

    def search(self, key: str) -> Search:
        """Search a collection.

        Tile38 Command:
            SEARCH key

        Args:
            key (str): The collection key.

        Returns:
            Search: An object to perform search queries.
        """
        return Search(self.client, key)

    async def server(self) -> ServerStatsResponseFollower:
        """Get server stats.

        Tile38 Command:
            SERVER

        Returns:
            ServerStatsResponseFollower: Basic server statistics.
        """
        return ServerStatsResponseFollower(
            **(await self.client.command(Command.SERVER))
        )

    async def server_extended(self) -> ServerStatsExtendedResponse:
        """Get extended server stats.

        Tile38 Command:
            SERVER

        Returns:
            ServerStatsExtendedResponse: Extended server statistics.
        """
        return ServerStatsExtendedResponse(
            **(await self.client.command(Command.SERVER, [SubCommand.EXT]))
        )

    async def stats(self, keys: list[str]) -> StatsResponse:
        """Get statistics for specified keys.

        Tile38 Command:
            STATS key...

        Args:
            keys (List[str]): A list of keys to get statistics for.

        Returns:
            StatsResponse: Statistics for the given keys.
        """
        response = await self.client.command(Command.STATS, keys)
        if response["stats"] == [None]:
            response["stats"] = []
        return StatsResponse(**response)

    async def quit(self) -> str:
        """Disconnect from Tile38.

        Tile38 Command:
            QUIT

        Returns:
            str: A confirmation message ("OK") upon successful termination.
        """
        await self.client.quit()
        return "OK"

    def within(self, key: str) -> Within:
        """Query objects within a specified area.

        Tile38 Command:
            WITHIN

        Args:
            key (str): The collection key.

        Returns:
            Within: An object to perform within-area queries.
        """
        return Within(self.client, key)



from collections.abc import Callable, Sequence
from enum import Enum
from typing import Any

from redis.asyncio import Connection, Redis
from redis.asyncio.connection import parse_url

from .client_options import ClientOptions
from .parse_response import parse_response

TILE38_DEFAULT_HOST = "localhost"
TILE38_DEFAULT_PORT = 9851


class Command(str, Enum):
    AOFSHRINK = "AOFSHRINK"
    AUTH = "AUTH"
    BOUNDS = "BOUNDS"
    CHANS = "CHANS"
    CONFIG = "CONFIG"
    DELCHAN = "DELCHAN"
    DEL = "DEL"
    DELHOOK = "DELHOOK"
    DROP = "DROP"
    EXISTS = "EXISTS"
    EXPIRE = "EXPIRE"
    FEXISTS = "FEXISTS"
    FLUSHDB = "FLUSHDB"
    FGET = "FGET"
    FSET = "FSET"
    GC = "GC"
    GET = "GET"
    HEALTHZ = "HEALTHZ"
    HOOKS = "HOOKS"
    INFO = "INFO"
    INTERSECTS = "INTERSECTS"
    JDEL = "JDEL"
    JGET = "JGET"
    JSET = "JSET"
    KEYS = "KEYS"
    NEARBY = "NEARBY"
    OUTPUT = "OUTPUT"
    PDELCHAN = "PDELCHAN"
    PDELHOOK = "PDELHOOK"
    PDEL = "PDEL"
    PERSIST = "PERSIST"
    PING = "PING"
    READONLY = "READONLY"
    RENAMENX = "RENAMENX"
    RENAME = "RENAME"
    SCAN = "SCAN"
    SEARCH = "SEARCH"
    SERVER = "SERVER"
    SETCHAN = "SETCHAN"
    SETHOOK = "SETHOOK"
    SET = "SET"
    STATS = "STATS"
    TTL = "TTL"
    WITHIN = "WITHIN"


class SubCommand(str, Enum):
    FIELD = "FIELD"
    EX = "EX"
    NX = "NX"
    XX = "XX"
    WITHFIELDS = "WITHFIELDS"
    CURSOR = "CURSOR"
    LIMIT = "LIMIT"
    NOFIELDS = "NOFIELDS"
    MATCH = "MATCH"
    OBJECT = "OBJECT"
    TILE = "TILE"
    QUADKEY = "QUADKEY"
    HASH = "HASH"
    BOUNDS = "BOUNDS"
    GET = "GET"
    CIRCLE = "CIRCLE"
    POINT = "POINT"
    OBJECTS = "OBJECTS"
    HASHES = "HASHES"
    IDS = "IDS"
    COUNT = "COUNT"
    POINTS = "POINTS"
    STRING = "STRING"
    META = "META"
    FENCE = "FENCE"
    DETECT = "DETECT"
    COMMANDS = "COMMANDS"
    REWRITE = "REWRITE"
    ASC = "ASC"
    DESC = "DESC"
    EXT = "EXT"
    WHERE = "WHERE"
    WHEREIN = "WHEREIN"
    SECTOR = "SECTOR"


class Format(str, Enum):
    JSON = "JSON"
    RESP = "RESP"


CommandArg = str | float | int
CommandArgs = Sequence[CommandArg] | Sequence[Sequence[CommandArg]]

NO_RESPONSE_CALLBACKS_FOR = [
    Command.SET,
    Command.SCAN,
    Command.RENAME,
    Command.RENAMENX,
    Command.PING,
    Command.PERSIST,
    Command.INFO,
    Command.EXPIRE,
    Command.FLUSHDB,
    Command.DEL,
    Command.READONLY,
]


class Client:
    """
    A class to manage the connection and communication with a Tile38 instance using Redis.

    Attributes:
        url (str): The connection URL for the Tile38 instance.
        __redis (Optional[redis.Redis]): A Redis connection object (initialized as None).
        __client_options (ClientOptions): Configuration options for the Redis client.
    """

    __redis = None
    __client_options: ClientOptions
    __url = ""

    def __init__(
        self, url: str, options: list[Callable[..., ClientOptions]] | None = None
    ) -> None:
        """Initialize the Client.

        Args:
            url (str): The URL for connecting to the Tile38 server.
            options (List[Callable[..., ClientOptions]], optional): A list of callables that modify client options.

        Returns:
            None
        """
        self.__url = url
        self.__redis = None
        self.__client_options = {}

        default_options: list[Callable[..., ClientOptions]] = []
        if options:
            default_options.extend(options)

        for option in default_options:
            self.__client_options = option(self.__client_options)

    @property
    def url(self) -> str:
        """The client url

        Returns:
            str: returns the clients url
        """
        return self.__url

    def client_options(self) -> ClientOptions:
        """Get the current ClientOptions.

        Returns:
            ClientOptions: The current configuration options for the client.
        """
        return self.__client_options

    async def __on_connect(self, connection: Connection) -> None:
        """Callback executed upon connection to Redis.

        Sets the OUTPUT format to JSON.

        Args:
            connection (redis.Connection): The Redis connection instance.

        Returns:
            None
        """
        await connection.on_connect()
        await connection.send_command(Command.OUTPUT, Format.JSON)
        await connection.read_response()  # pyright: ignore[reportUnusedCallResult]

    async def __delete_response_callbacks(self) -> None:
        """Delete unnecessary response callbacks in redis-py.

        This removes default callbacks for certain commands that cause issues with Tile38.
        These commands are specified in NO_RESPONSE_CALLBACKS_FOR.

        Returns:
            None
        """
        r = await self.__get_redis()
        for command in NO_RESPONSE_CALLBACKS_FOR:
            try:
                del r.response_callbacks[command]
            except KeyError:
                continue

    async def __get_redis(self) -> Redis:
        """Establish or retrieve the Redis connection.

        This method acts as a singleton for the Redis connection, ensuring that only one connection
        is maintained. It sets up the connection with specified host, port, and client options.

        Returns:
            redis.Redis: The Redis connection instance.
        """
        if not self.__redis:
            url_components = parse_url(self.__url)
            host: str = url_components.get("host") or TILE38_DEFAULT_HOST
            port: int = url_components.get("port") or TILE38_DEFAULT_PORT

            r: Redis = Redis(
                host=host,
                port=port,
                encoding="utf-8",
                single_connection_client=True,
                decode_responses=True,
                redis_connect_func=self.__on_connect,
                **self.__client_options,
            )

            self.__redis = r

            # Remove unnecessary response callbacks specific to Tile38
            await self.__delete_response_callbacks()

        return self.__redis

    async def __execute_and_read_response(  # pyright: ignore[reportUnknownParameterType]
        self, command: str, command_args: CommandArgs | None
    ) -> Any:
        """Execute a command and read its response from Redis.

        Args:
            command (str): The command to execute.
            command_args (CommandArgs, optional): Arguments for the command.

        Returns:
            Any: The response from the Redis server.
        """
        command_args = command_args if command_args else []
        r = await self.__get_redis()
        return await r.execute_command(command, *command_args)

    async def command(
        self, command: str, command_args: CommandArgs | None = None
    ) -> dict:
        """Send a command to Tile38 and parse the response.

        Args:
            command (str): The command to execute on the Tile38 server.
            command_args (CommandArgs, optional): Arguments for the command.

        Returns:
            Dict: The parsed response from the server.
        """
        if not command_args:
            command_args = []

        response = await self.__execute_and_read_response(command, command_args)
        return parse_response(response)

    async def quit(self) -> str:
        """Terminate the Redis connection.

        Closes the connection to the Tile38 instance and resets the Redis client.

        Returns:
            str: A confirmation message ("OK") upon successful termination.
        """
        if not self.__redis:
            return "OK"

        c = await self.__get_redis()

        # TODO: remove ignore once types have been updated
        await c.aclose()  # type: ignore[attr-defined]
        await c.connection_pool.disconnect()

        self.__redis = None

        return "OK"



import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38KeyNotFoundError

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_del(tile38: Tile38) -> None:
    key = random_string()
    oid = random_string()

    response = await tile38.set(key, oid).point(1, 1).exec()
    assert response.ok

    response = await tile38.delete(key, oid)
    assert response.ok

    with pytest.raises(Tile38KeyNotFoundError):
        await tile38.get(key, oid).asObject()



import pytest

from pyle38 import Tile38

from .helper.random_data import random_polygon_feature, random_string


@pytest.mark.asyncio
async def test_command_hooks(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    key = random_string()
    oid = random_string()
    obj = random_polygon_feature()
    endpoint = "kafka://10.0.20.78:9092/warehouse"

    response = await tile38.sethook(oid, endpoint).within(key).object(obj).activate()

    assert response.ok

    response = await tile38.hooks()
    assert response.ok
    assert len(response.hooks) == 1



from typing import Any

import pytest

from pyle38 import Tile38
from pyle38.commands.executable import Compiled
from pyle38.commands.within import Format, Within
from pyle38.models import Feature, Polygon

from .helper.random_data import random_string

key = random_string()
oid = random_string()
feature = {
    "type": "Feature",
    "geometry": {"type": "Point", "coordinates": [13.37, 52.25]},
    "properties": {"id": oid},
}
polygon = {
    "type": "Polygon",
    "coordinates": [
        [
            [13.361263275146484, 52.24630137198303],
            [13.379974365234373, 52.24630137198303],
            [13.379974365234373, 52.256705331409506],
            [13.361263275146484, 52.256705331409506],
            [13.361263275146484, 52.24630137198303],
        ]
    ],
}

expected = {"id": oid, "object": feature}


@pytest.mark.parametrize(
    "fmt, expected",
    [
        (
            "OBJECTS",
            [
                "WITHIN",
                [
                    key,
                    "MATCH",
                    "*",
                    "BUFFER",
                    10,
                    "NOFIELDS",
                    "SPARSE",
                    1,
                    "CURSOR",
                    0,
                    "LIMIT",
                    10,
                    "WHERE",
                    "foo",
                    1,
                    1,
                    "WHERE",
                    "bar",
                    1,
                    1,
                    "WHEREIN",
                    "foo",
                    1,
                    1,
                    "WHEREIN",
                    "bar",
                    1,
                    1,
                    "FENCE",
                    "DETECT",
                    "enter,exit",
                    "COMMANDS",
                    "del,set",
                    "CIRCLE",
                    1.0,
                    1.0,
                    100,
                ],
            ],
        )
    ],
    ids=["OBJECTS"],
)
@pytest.mark.asyncio
async def test_command_within_compile(
    tile38: Tile38, fmt: Format, expected: Compiled
) -> None:
    query = (
        Within(tile38.client, key)
        .match("*")
        .buffer(10)
        .nofields()
        .sparse(1)
        .cursor(0)
        .limit(10)
        .where("foo", 1, 1)
        .where("bar", 1, 1)
        .wherein("foo", [1])
        .wherein("bar", [1])
        .fence()
        .detect(["enter", "exit"])
        .commands(["del", "set"])
        .circle(1, 1, 100)
    )

    received = query.output(fmt).compile()

    assert expected == received


@pytest.mark.asyncio
async def test_command_within_circle(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_where_circle(tile38: Tile38) -> None:
    await tile38.set(key, oid).fields({"maxspeed": 120}).object(feature).exec()
    await tile38.set(key, "truck2").fields({"maxspeed": 100}).object(feature).exec()

    response = (
        await tile38
        .within(key)
        .where("maxspeed", 120, 120)
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120]})

    response = (
        await tile38
        .within(key)
        .where("maxspeed", 100, 120)
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_within_wherein_circle(tile38: Tile38) -> None:
    await tile38.set(key, oid).fields({"maxspeed": 120}).object(feature).exec()
    await tile38.set(key, "truck2").fields({"maxspeed": 100}).object(feature).exec()

    response = (
        await tile38
        .within(key)
        .wherein("maxspeed", [120])
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 1
    assert response.objects[0].dict() == dict(expected, **{"fields": [120]})

    response = (
        await tile38
        .within(key)
        .wherein("maxspeed", [100, 120])
        .circle(52.25, 13.37, 100)
        .asObjects()
    )
    assert response.ok
    assert len(response.objects) == 2


@pytest.mark.asyncio
async def test_command_within_object(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    polygon: dict[str, Any] = {
        "type": "Polygon",
        "coordinates": [
            [
                [13.361263275146484, 52.24630137198303],
                [13.379974365234373, 52.24630137198303],
                [13.379974365234373, 52.256705331409506],
                [13.361263275146484, 52.256705331409506],
                [13.361263275146484, 52.24630137198303],
            ]
        ],
    }

    response = await tile38.within(key).object(polygon).asObjects()

    assert response.ok
    assert response.objects[0].dict() == expected

    response = await tile38.within(key).object(Polygon(**polygon)).asObjects()

    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_hash(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).hash("u3390").asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_quadkey(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).quadkey("120").asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_tile(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).tile(2200, 1348, 12).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).bounds(52.24, 13.36, 52.256, 13.379).asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_sector(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = (
        await tile38.within(key).sector(52.25191, 13.37230, 1000, 180, 270).asObjects()
    )
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_get(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.set("zones", "zone").object(polygon).exec()
    assert response.ok

    response = await tile38.within(key).get("zones", "zone").asObjects()
    assert response.ok
    assert response.objects[0].dict() == expected


@pytest.mark.asyncio
async def test_command_within_return_points(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asPoints()
    assert response.ok
    assert response.points[0].dict() == {
        "id": oid,
        "point": {"lat": 52.25, "lon": 13.37},
    }


@pytest.mark.asyncio
async def test_command_within_return_ids(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asIds()
    assert response.ok
    assert response.ids == [oid]


@pytest.mark.asyncio
async def test_command_within_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asCount()
    assert response.ok
    assert response.count == 1


@pytest.mark.asyncio
async def test_command_within_return_hashes(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asHashes(5)
    assert response.ok
    assert response.hashes[0].dict() == {"id": oid, "hash": "u3390"}


@pytest.mark.asyncio
async def test_command_within_return_bounds(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    response = await tile38.within(key).circle(52.25, 13.37, 100).asBounds()
    assert response.ok
    assert response.bounds[0].dict() == {
        "id": oid,
        "bounds": {
            "ne": {"lat": 52.25, "lon": 13.37},
            "sw": {"lat": 52.25, "lon": 13.37},
        },
    }


@pytest.mark.asyncio
async def test_command_within_buffer_return_count(tile38: Tile38) -> None:
    response = await tile38.set(key, oid).object(feature).exec()
    assert response.ok

    search_area: dict[str, Any] = {
        "type": "Feature",
        "properties": {},
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [13.37009847164154, 52.2498254610514],
                    [13.370516896247862, 52.2498254610514],
                    [13.370516896247862, 52.25017851139772],
                    [13.37009847164154, 52.25017851139772],
                    [13.37009847164154, 52.2498254610514],
                ]
            ],
        },
    }

    response = await tile38.within(key).object(search_area).asCount()
    assert response.ok
    assert response.count == 0

    response = await tile38.within(key).buffer(10).object(search_area).asCount()
    assert response.ok
    assert response.count == 1

    response = (
        await tile38.within(key).buffer(10).object(Feature(**search_area)).asCount()
    )
    assert response.ok
    assert response.count == 1



from collections.abc import Callable
from typing import TypeAlias

from redis.asyncio.retry import Retry
from redis.backoff import ExponentialBackoff
from typing_extensions import NotRequired, TypedDict

from .errors import Pyle38Error

Pyle38Retry: TypeAlias = Retry
Pyle38ExponentialBackoff: TypeAlias = ExponentialBackoff


class ClientOptions(TypedDict):
    retry: NotRequired[Pyle38Retry]
    retry_on_error: NotRequired[list[type[Pyle38Error]]]


def WithRetryExponentialBackoff(retries: int) -> Callable[..., ClientOptions]:
    """Return a callable that configures exponential backoff for retries.

    This function creates and returns a callable that, when invoked, will configure
    retry behavior with exponential backoff for the given number of retries.

    Args:
        retries (int): The number of retry attempts to make.

    Raises:
        ValueError: If the retries argument is not a positive integer.

    Returns:
        Callable[..., ClientOptions]: A function that accepts `ClientOptions` and
        configures exponential backoff retry with the specified number of retries.
    """
    try:
        t = int(retries)
    except ValueError as e:
        raise ValueError("Retries must be a positive integer") from e  # noqa: TRY003

    if t < 0:
        raise ValueError("Retries must be a positive integer")  # noqa: TRY003

    def _with_retry_exponential_backoff(opts: ClientOptions) -> ClientOptions:
        """Helper function to update options with retry strategy.

        This function modifies the `ClientOptions` by setting the `retry` field
        to the configured exponential backoff retry strategy.

        Args:
            opts (ClientOptions): The current client options to update.

        Returns:
            ClientOptions: The updated client options with retry configuration.
        """
        opts["retry"] = Pyle38Retry(Pyle38ExponentialBackoff(), retries=retries)
        return opts

    return _with_retry_exponential_backoff


def WithRetryOnError(
    *errs: type[Pyle38Error],
) -> Callable[..., ClientOptions]:
    """Return a callable that configures retry behavior based on specific errors.

    This function creates and returns a callable that, when invoked, will configure
    the retry behavior to only retry on the specified error types.

    Args:
        *errs (type[Pyle38Error]): One or more error types to retry on.

    Returns:
        Callable[..., ClientOptions]: A function that accepts `ClientOptions` and
        configures the retry behavior based on the specified error types.
    """

    def _with_retry_on_error(opts: ClientOptions) -> ClientOptions:
        """Helper function to update options with error-based retry configuration.

        This function modifies the `ClientOptions` by setting the `retry_on_error`
        field to the provided list of error types that should trigger retries.

        Args:
            opts (ClientOptions): The current client options to update.

        Returns:
            ClientOptions: The updated client options with error-based retry configuration.
        """
        opts["retry_on_error"] = []
        if len(errs) > 0:
            opts["retry_on_error"] = list(errs)
        return opts

    return _with_retry_on_error



import os
from collections.abc import Callable

from pyle38.client_options import ClientOptions

from .errors import Pyle38NoFollowerSetError, Pyle38NoLeaderSetError
from .follower import Follower
from .leader import Leader

TILE38_LEADER_URI: str | None = os.getenv("TILE38_LEADER_URI")
TILE38_FOLLOWER_URI: str | None = os.getenv("TILE38_FOLLOWER_URI")


class Tile38(Leader):
    """
    A class to manage a Tile38 leader instance with optional follower support.

    Attributes:
        __follower (Optional[Follower]): A private attribute holding the follower instance if a follower URL is provided.
    """

    __follower: Follower | None = None

    def __init__(
        self,
        url: str | None = TILE38_LEADER_URI,
        follower_url: str | None = TILE38_FOLLOWER_URI,
        options: list[Callable[..., ClientOptions]] = [],
    ) -> None:
        """Initialize the Tile38 leader and optional follower.

        Args:
            url (str, optional): The URL for the Tile38 leader instance.
                Defaults to the environment variable `TILE38_LEADER_URI`.
            follower_url (str, optional): The URL for the Tile38 follower instance.
                Defaults to the environment variable `TILE38_FOLLOWER_URI`.
            options (List[Callable[..., ClientOptions]], optional): A list of callables
                that return client options for configuring the Tile38 connection.

        Raises:
            Tile38Error: If the leader URL is not provided.

        Returns:
            None
        """
        if not url:
            raise Pyle38NoLeaderSetError

        super().__init__(url, options)

        if follower_url:
            self.__follower = Follower(follower_url, options)

    def follower(self) -> Follower:
        """Get the Tile38 follower instance.

        Raises:
            Tile38Error: If the follower instance is not available.

        Returns:
            Follower: The follower instance.
        """
        if not self.__follower:
            raise Pyle38NoFollowerSetError

        return self.__follower

    async def quit(self) -> str:
        """Asynchronously quit the Tile38 leader and follower connections.

        This method first quits the follower connection (if available),
        followed by quitting the leader connection.

        Returns:
            str: A confirmation message ("OK") upon successful termination.
        """
        if self.__follower:
            await self.follower().quit()

        await super().quit()

        return "OK"



import pytest

from pyle38 import Tile38
from pyle38.errors import Tile38KeyNotFoundError, Tile38PathNotFoundError

from .helper.random_data import random_string


@pytest.mark.asyncio
async def test_command_jset_jget_jdel(tile38_with_follower: Tile38) -> None:
    tile38 = tile38_with_follower

    response = await tile38.jset(key="user", oid="901", path="name", value="Tom")
    assert response.ok

    response = await tile38.jget("user", "901")
    assert response.ok
    assert response.value == '{"name":"Tom"}'

    response = await tile38.jdel("user", "901", "name")
    assert response.ok

    response = await tile38.jget("user", "901", "name")
    assert response.ok
    assert response.value == "{}"


@pytest.mark.asyncio
async def test_command_j_options(tile38: Tile38) -> None:
    obj = {"type": "LineString", "coordinates": [[0, 0], [1, 1]]}

    response = await tile38.set("linestring", "1").object(obj).exec()
    assert response.ok

    response = await tile38.get("linestring", "1").asObject()
    assert response.ok
    assert response.object == obj

    response = await tile38.jset("linestring", "1", "coordinates.-1", "[2,2]", "RAW")
    assert response.ok

    response = await tile38.jget("linestring", "1")
    assert response.ok
    assert response.value == '{"type":"LineString","coordinates":[[0,0],[1,1],[2,2]]}'

    with pytest.raises(Tile38KeyNotFoundError):
        await tile38.jdel(random_string(), random_string(), random_string())

    with pytest.raises(Tile38PathNotFoundError):
        await tile38.jdel("linestring", random_string(), random_string())



import pytest

from pyle38 import Tile38
from pyle38.commands.sethook import SetHook
from pyle38.errors import Pyle38NoHookToActivateError

from .helper.random_data import random_string

endpoint = "kafka://10.0.20.78:9092/warehouse"


@pytest.mark.asyncio
async def test_command_sethook_compile_within(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()
    meta_key = random_string()
    meta_value = random_string()

    received = (
        SetHook(tile38.client, name, endpoint)
        .meta({meta_key: meta_value})
        .ex(10)
        .within(key)
        .circle(52.25, 13.37, 100)
        .compile()
    )

    assert received == [
        "SETHOOK",
        [
            name,
            endpoint,
            "META",
            meta_key,
            meta_value,
            "EX",
            10,
            "WITHIN",
            key,
            "FENCE",
            "CIRCLE",
            52.25,
            13.37,
            100.0,
        ],
    ]


@pytest.mark.asyncio
async def test_command_Sethook_compile_nearby(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()

    received = (
        SetHook(tile38.client, name, endpoint)
        .intersects(key)
        .circle(52.25, 13.37, 100)
        .compile()
    )

    assert received == [
        "SETHOOK",
        [name, endpoint, "INTERSECTS", key, "FENCE", "CIRCLE", 52.25, 13.37, 100.0],
    ]


@pytest.mark.asyncio
async def test_command_sethook_compile_nearby(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()

    received = (
        SetHook(tile38.client, name, endpoint)
        .nearby(key)
        .point(52.25, 13.37, 100)
        .compile()
    )

    assert received == [
        "SETHOOK",
        [name, endpoint, "NEARBY", key, "FENCE", "POINT", 52.25, 13.37, 100.0],
    ]


@pytest.mark.asyncio
async def test_command_sethook(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()

    response = (
        await tile38
        .sethook(name, endpoint)
        .within(key)
        .circle(52.25, 13.37, 100)
        .activate()
    )
    assert response.ok

    response = (
        await tile38
        .sethook(name, endpoint)
        .intersects(key)
        .circle(52.25, 13.37, 100)
        .activate()
    )
    assert response.ok

    response = (
        await tile38
        .sethook(name, endpoint)
        .nearby(key)
        .point(52.25, 13.37, 100)
        .activate()
    )
    assert response.ok


@pytest.mark.asyncio
async def test_command_sethook_raises(tile38: Tile38) -> None:
    key = random_string()

    with pytest.raises(Pyle38NoHookToActivateError):
        await tile38.within(key).circle(52.25, 13.37, 100).activate()

    with pytest.raises(Pyle38NoHookToActivateError):
        await tile38.intersects(key).circle(52.25, 13.37, 100).activate()

    with pytest.raises(Pyle38NoHookToActivateError):
        await tile38.nearby(key).point(52.25, 13.37, 100).activate()


@pytest.mark.asyncio
async def test_command_pdelhook(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()

    response = (
        await tile38
        .sethook(name, endpoint)
        .within(key)
        .circle(52.25, 13.37, 100)
        .activate()
    )
    assert response.ok

    response = await tile38.hooks()
    assert response.ok
    assert len(response.hooks) == 1

    response = await tile38.pdelhook("*")
    assert response.ok

    response = await tile38.hooks()
    assert response.ok
    assert len(response.hooks) == 0


@pytest.mark.asyncio
async def test_command_delhook(tile38: Tile38) -> None:
    key = random_string()
    name = random_string()

    response = (
        await tile38
        .sethook(name, endpoint)
        .within(key)
        .circle(52.25, 13.37, 100)
        .activate()
    )
    assert response.ok

    response = await tile38.hooks()
    assert response.ok
    assert len(response.hooks) == 1

    response = await tile38.delhook(name)
    assert response.ok

    response = await tile38.hooks()
    assert response.ok
    assert len(response.hooks) == 0
